from .arrays import Array
from .types import Any, Callable, ShapeError, arr, config

MatchError = ShapeError("The shapes of two matrices do not match")
SquareError = ShapeError("Matrix must be square")


class Matrix(Array):
    """
    It is a matrix class that supports basic operations,
    as well as functions such as determinant and inverse matrix.
    :param data: An array in the form of a list
    :param check: Check the rationality of the input array
    """

    def __init__(self, data, check=True):
        super().__init__(data, check)
        if len(self.shape) != 2:
            raise ShapeError("Matrix can only be two-dimensional in shape")
        self.rows = len(data)
        self.cols = len(data[0])

    def __add__(self, other):
        if self.rows != other.rows or self.cols != other.cols:
            raise MatchError
        return Matrix([[self.data[i][j] + other.data[i][j] for j in range(self.cols)] for i in range(self.rows)])

    def __sub__(self, other):
        if self.rows != other.rows or self.cols != other.cols:
            raise MatchError
        return Matrix([[self.data[i][j] - other.data[i][j] for j in range(self.cols)] for i in range(self.rows)])

    def __mul__(self, other):
        if isinstance(other, Matrix):
            if self.rows != other.rows or self.cols != other.cols:
                raise MatchError
            return Matrix([[self.data[i][j] * other.data[i][j] for j in range(self.cols)] for i in range(self.rows)])
        else:
            return Matrix([[item * other for item in row] for row in self.data])

    def __matmul__(self, other):
        if self.cols != other.rows:
            raise MatchError
        return Matrix([[sum([self.data[i][k] * other.data[k][j] for k in range(self.cols)]) for j in range(other.cols)]
                       for i in range(self.rows)])

    def kron(self, other):
        return Matrix([[self.data[i][j] * other.data[m][n] for j in range(self.cols) for n in range(other.cols)]
                       for i in range(self.rows) for m in range(other.rows)])

    def inner(self, other):
        if self.cols != other.cols:
            raise MatchError
        return Matrix([[sum([self.data[i][k] * other.data[j][k] for k in range(self.cols)]) for j in range(other.rows)]
                       for i in range(self.rows)])

    def outer(self, other):
        vec1, vec2 = self.flatten(), other.flatten()
        return Matrix([[a * b for b in vec2] for a in vec1])

    def t(self):
        return Matrix(list(map(list, zip(*self.data))))

    def minor(self, row, col):
        return Matrix([row[:col] + row[col + 1:] for row in self.data[:row] + self.data[row + 1:]])

    def tr(self):
        if self.rows != self.cols:
            raise SquareError
        return sum([self.data[i][i] for i in range(self.rows)])

    def det(self):
        if self.rows != self.cols:
            raise SquareError
        matrix = self.copy().data
        size = len(matrix)
        determinant = 1
        for i in range(size):
            if matrix[i][i] == 0:
                for j in range(i + 1, size):
                    if matrix[i][j] != 0:
                        matrix[i], matrix[j] = matrix[j], matrix[i]
                        determinant = -determinant
                        break
                if matrix[i][i] == 0:
                    return 0
            determinant *= matrix[i][i]
            for j in range(i + 1, size):
                matrix[i][j] /= matrix[i][i]
                for k in range(i + 1, size):
                    matrix[k][j] -= matrix[i][j] * matrix[k][i]
        return determinant

    def slogdet(self):
        from math import log
        det_value = self.det()
        return (0.0, float("-inf")) if det_value == 0 else (det_value / abs(det_value), log(abs(det_value)))

    def inv(self):
        if self.rows != self.cols:
            raise SquareError
        n = self.rows
        augmented = [self[i, :] + [0 if k != i else 1 for k in range(n)] for i in range(n)]
        for i in range(n):
            max_el = abs(augmented[i][i])
            max_row = i
            for k in range(i + 1, n):
                if abs(augmented[k][i]) > max_el:
                    max_el = abs(augmented[k][i])
                    max_row = k
            if max_el == 0:
                raise ValueError("Matrix is singular and cannot be inverted")
            if max_row != i:
                augmented[i], augmented[max_row] = augmented[max_row], augmented[i]
            pivot = augmented[i][i]
            for j in range(n * 2):
                augmented[i][j] /= pivot
            for j in range(n):
                if j != i:
                    factor = augmented[j][i]
                    for k in range(n * 2):
                        augmented[j][k] -= factor * augmented[i][k]
        return Matrix([augmented[i][n:] for i in range(n)])

    def rref(self, pivots=True):
        pe = []
        matrix = self.copy().data
        lead = 0
        for r in range(self.rows):
            if lead >= self.cols:
                return (Matrix(matrix), pe) if pivots else Matrix(matrix)
            i = r
            while matrix[i][lead] == 0:
                i += 1
                if i == self.rows:
                    i = r
                    lead += 1
                    if self.cols == lead:
                        return (Matrix(matrix), pe) if pivots else Matrix(matrix)
            matrix[i], matrix[r] = matrix[r], matrix[i]
            pe.append(lead)
            lv = matrix[r][lead]
            matrix[r] = [mrx / lv for mrx in matrix[r]]
            for i in range(self.rows):
                if i != r:
                    lv = matrix[i][lead]
                    matrix[i] = [iv - lv * rv for rv, iv in zip(matrix[r], matrix[i])]
            lead += 1
        return (Matrix(matrix), pe) if pivots else Matrix(matrix)

    def rank(self):
        matrix = self.rref(False).data
        for _ in range(self.rows):
            if matrix[_].count(0) == self.cols:
                return _
        return self.rows

    def diag(self):
        if self.cols > 1:
            return Matrix([[self.data[i][i]] for i in range(self.rows)])
        else:
            result = [[0] * self.rows for _ in range(self.rows)]
            for i in range(self.rows):
                result[i][i] = self.data[i][0]
            return Matrix(result)

    def i(self):
        return identity(self.rows, self.cols)

    def norm(self, p=None):
        if p is None:
            return sum([sum([item * item for item in row]) for row in self.data]) ** 0.5
        elif p == 1:
            return max([sum([abs(item) for item in row]) for row in self.t().data])
        elif p == 2:
            return svd(self)[1][0, 0]
        elif p == float("inf"):
            return max([sum([abs(item) for item in row]) for row in self.data])
        else:
            raise ValueError("Invalid norm order of matrix")

    def normalize(self):
        norm = self.norm()
        return Matrix([[item / norm for item in row] for row in self.data])

    determinant = det
    diagonal = diag
    identity = i
    inverse = inv
    transpose = t
    trace = tr

    def latex(self, matrix_type="bmatrix", row_spacing=0, col_spacing=0, spacing_unit="mm"):
        valid_matrix_types = ["Bmatrix", "Vmatrix", "array", "bmatrix", "matrix", "pmatrix", "smallmatrix", "vmatrix"]
        valid_units = ["bp", "cc", "cm", "dd", "em", "ex", "in", "mm", "mu", "pc", "pt", "sp"]
        if matrix_type not in valid_matrix_types:
            raise ValueError("Invalid matrix type. Choose from {}".format(valid_matrix_types))
        if spacing_unit not in valid_units:
            raise ValueError("Invalid spacing unit. Choose from {}".format(valid_units))
        alignment = "c" * len(self[0]) if matrix_type == "array" else None
        col_separator = " & " if col_spacing == 0 else " & \\hspace{{{}{}}} & ".format(col_spacing, spacing_unit)
        row_separator = "\\\\" if row_spacing == 0 else "\\\\[{}{}]".format(row_spacing, spacing_unit)
        latex_str = ["\\begin{{{}}}".format(matrix_type), "{{{}}}".format(alignment) if alignment else "",
                     row_separator.join([col_separator.join(map(str, row)) for row in self]),
                     "\\end{{{}}}".format(matrix_type)]
        return "".join(latex_str)

    def __str__(self, use_latex=False):
        use_latex = config.use_latex or use_latex
        return self.latex() if use_latex else super().__str__()

    def __setitem__(self, key, value):
        if isinstance(key, (int, slice)):
            self.data[key] = value
        elif isinstance(key, tuple):
            if len(key) == 2:
                if isinstance(key[0], int) and isinstance(key[1], (int, slice)):
                    self.data[key[0]][key[1]] = value
                elif isinstance(key[0], slice) and isinstance(key[1], (int, slice)):
                    for old, new in zip(self.data[key[0]], value[key[0]]):
                        old[key[1]] = new[key[1]]
                else:
                    raise TypeError("The type of index is invalid")
            else:
                raise ValueError("The length of the index can only be one or two")
        else:
            raise TypeError("The type of index is invalid")


def mat(data):
    return Matrix(data)


def rotate90(matrix: Matrix, times: int) -> Matrix:
    matrix = matrix.copy()
    if times % 4 == 0:
        return matrix
    elif times % 4 == 1:
        return Matrix([list(col)[::-1] for col in zip(*matrix)])
    elif times % 4 == 2:
        return Matrix([row[::-1] for row in matrix[::-1]])
    else:
        return Matrix([list(col)[::-1] for col in zip(*matrix[::-1])][::-1])


def rank_decomp(matrix: Matrix) -> tuple:
    f, pivot_cols = matrix.rref()
    rank = len(pivot_cols)
    c = matrix[range(matrix.rows), pivot_cols]
    f = f[:rank, :]
    return Matrix(c, False), Matrix(f, False)


def cholesky(matrix: Matrix, hermitian: bool = True) -> Matrix:
    rows = matrix.rows
    cols = matrix.cols
    if rows != cols:
        raise SquareError
    lower = [[0.0] * cols for _ in range(rows)]
    if hermitian:
        for i in range(matrix.rows):
            for j in range(i):
                lower[i][j] = (matrix[i][j] - sum([lower[i][k] * lower[j][k].conjugate()
                                                   for k in range(j)])) / lower[j][j]
            lower[i][i] = (matrix[i][i] - sum([lower[i][k] * lower[i][k].conjugate() for k in range(i)])) ** 0.5
    else:
        for i in range(matrix.rows):
            for j in range(i):
                lower[i][j] = (matrix[i][j] - sum([lower[i][k] * lower[j][k] for k in range(j)])) / lower[j][j]
            lower[i][i] = (matrix[i][i] - sum([lower[i][k] ** 2 for k in range(i)])) ** 0.5
    return Matrix(lower, False)


def lu(matrix: Matrix) -> tuple:
    rows = matrix.rows
    cols = matrix.cols
    if rows != cols:
        raise SquareError
    lower = [[0.0] * cols for _ in range(rows)]
    for i in range(rows):
        lower[i][i] = 1.0
    upper = matrix[:, :]
    for i in range(rows - 1):
        for j in range(i + 1, rows):
            lower[j][i] = upper[j][i] / upper[i][i]
        for j in range(i + 1, rows):
            for k in range(i + 1, rows):
                upper[j][k] -= lower[j][i] * upper[i][k]
        for j in range(i + 1, rows):
            upper[j][i] = 0.0
    return Matrix(lower), Matrix(upper)


def triu_indices(n: int, k: int = 0, m: int = None) -> tuple:
    if m is None:
        m = n
    if n < 1 or k >= m or m < 1:
        return (), ()
    indices = [(y, x) for y in range(n) for x in range(y + k, m) if x > -1]
    return tuple(zip(*indices))


def tril_indices(n: int, k: int = 0, m: int = None) -> tuple:
    if m is None:
        m = n
    if n < 1 or k <= -n or m < 1:
        return (), ()
    indices = [(y, x) for y in range(n) for x in range(y + k + 1) if x < m]
    return tuple(zip(*indices))


def identity(n: int, m: int = None) -> Matrix:
    if m is None:
        m = n
    result = [[0] * m for _ in range(n)]
    minimum = min(n, m)
    for i in range(minimum):
        result[i][i] = 1
    return Matrix(result)


def perm_mat_indices(num_rows: int, num_cols: int, row_swaps: arr = (), col_swaps: arr = ()) -> tuple:
    if not (isinstance(num_rows, int) and num_rows > 0 and isinstance(num_cols, int) and num_cols > 0):
        raise ValueError("Number of rows and columns must be positive integers")
    if set(map(len, row_swaps)) != {2} or set(map(len, col_swaps)) != {2}:
        raise ValueError("Swap information must be tuples of two elements")
    row_coords = list(range(num_rows))
    col_coords = list(range(num_cols))

    def adjust_index(index, limit):
        return index + limit if index < 0 else index

    for r1, r2 in row_swaps:
        r1, r2 = adjust_index(r1, num_rows), adjust_index(r2, num_rows)
        if not (0 <= r1 < num_rows and 0 <= r2 < num_rows):
            raise ValueError("Row index out of range")
        row_coords[r1], row_coords[r2] = row_coords[r2], row_coords[r1]
    for c1, c2 in col_swaps:
        c1, c2 = adjust_index(c1, num_cols), adjust_index(c2, num_cols)
        if not (0 <= c1 < num_cols and 0 <= c2 < num_cols):
            raise ValueError("Column index out of range")
        col_coords[c1], col_coords[c2] = col_coords[c2], col_coords[c1]
    return row_coords, col_coords


def perm_mat(num_rows: int, num_cols: int, row_swaps: arr = (), col_swaps: arr = (), rtype: Callable = Matrix) -> Any:
    row_coords, col_coords = perm_mat_indices(num_rows, num_cols, row_swaps, col_swaps)
    result = [[0] * num_cols for _ in range(num_rows)]
    for i in range(num_rows):
        result[i][col_coords[row_coords[i]]] = 1
    return result if rtype is list else rtype(result)


def qr(matrix: Matrix, reduce: bool = False) -> tuple:
    from math import hypot
    r_ = matrix.rows
    c_ = matrix.cols
    q = identity(r_)
    r = matrix
    rows, cols = tril_indices(r_, -1, c_)
    for row, col in zip(rows, cols):
        if r[row][col] != 0:
            _r = hypot(r[col][col], r[row][col])
            _c = r[col][col] / _r
            _s = -r[row][col] / _r
            g = identity(r_)
            g[col][col] = _c
            g[row][row] = _c
            g[row][col] = _s
            g[col][row] = -_s
            q = q.inner(g)
            r = g @ r
    if reduce:
        minimum = min(r_, c_)
        q = Matrix(q[:, :minimum], False)
        r = Matrix(r[:minimum], False)
    return q, r


def hessenberg(matrix: Matrix) -> tuple:
    from math import sqrt
    a = matrix.copy()
    n = a.rows
    t = [0.0] * n
    if isinstance(a[-1, -1], int):
        a[-1, -1] = float(a[-1, -1])
    if n > 2:
        inf = float("inf")
        for i in range(n - 1, 1, -1):
            scale = 0.0
            for k in range(0, i):
                scale += abs(a[i, k].real) + abs(a[i, k].imag)
            scale_inv = 0.0
            if scale != 0:
                scale_inv = 1 / scale
            if scale == 0 or scale_inv == inf:
                t[i] = 0.0
                a[i, i - 1] = 0.0
                continue
            h = 0.0
            for k in range(0, i):
                a[i, k] *= scale_inv
                rr, ii = a[i, k].real, a[i, k].imag
                h += rr * rr + ii * ii
            f = a[i, i - 1]
            f0 = abs(f)
            g = sqrt(h)
            a[i, i - 1] = - g * scale
            if f0 == 0:
                t[i] = g
            else:
                ff = f / f0
                t[i] = f + g * ff
                a[i, i - 1] = a[i, i - 1] * ff
            h += g * f0
            h = 1 / sqrt(h)
            t[i] *= h
            for k in range(0, i - 1):
                a[i, k] *= h
            for j in range(0, i):
                g = t[i].conjugate() * a[j, i - 1]
                for k in range(0, i - 1):
                    g += a[i, k].conjugate() * a[j, k]
                a[j, i - 1] = a[j, i - 1] - g * t[i]
                for k in range(0, i - 1):
                    a[j, k] = a[j, k] - g * a[i, k]
            for j in range(0, n):
                g = t[i] * a[i - 1, j]
                for k in range(0, i - 1):
                    g += a[i, k] * a[k, j]
                a[i - 1, j] = a[i - 1, j] - g * t[i].conjugate()
                for k in range(0, i - 1):
                    a[k, j] = a[k, j] - g * a[i, k].conjugate()
    q = a.copy()
    if n == 1:
        q[0, 0] = 1.0
        return q, a
    q[0, 0] = q[1, 1] = 1.0
    q[0, 1] = q[1, 0] = 0.0
    for i in range(2, n):
        if t[i] != 0:
            for j in range(0, i):
                g = t[i] * q[i - 1, j]
                for k in range(0, i - 1):
                    g += q[i, k] * q[k, j]
                q[i - 1, j] -= g * t[i].conjugate()
                for k in range(0, i - 1):
                    q[k, j] -= g * a[i, k].conjugate()
        q[i, i] = 1.0
        for j in range(0, i):
            q[j, i] = q[i, j] = 0.0
    for x in range(n):
        for y in range(x + 2, n):
            a[y, x] = 0.0
    return q, a


def eigen(matrix: Matrix) -> tuple:
    from math import hypot
    from cmath import sqrt
    a = matrix.copy()
    n = a.rows
    if n == 1:
        return Matrix([[float(a[0, 0])]]), Matrix([[1.0]])
    q, a = hessenberg(a)
    norm = 0
    for x in range(n):
        for y in range(min(x + 2, n)):
            norm += abs(a[y, x])
    norm = norm ** 0.5 / n
    if norm == 0:
        return None, None
    n0 = 0
    n1 = n
    its = 0
    while True:
        k = n0
        while k + 1 < n1:
            if abs(a[k + 1, k]) <= 1e-100:
                break
            k += 1
        if k + 1 < n1:
            a[k + 1, k] = 0.0
            n0 = k + 1
            its = 0
            if n0 + 1 >= n1:
                n0 = 0
                n1 = k + 1
                if n1 < 2:
                    break
        else:
            if its % 30 == 10:
                shift = a[n1 - 1, n1 - 2]
            elif its % 30 == 20:
                shift = abs(a[n1 - 1, n1 - 2])
            elif its % 30 == 29:
                shift = norm
            else:
                t = a[n1 - 2, n1 - 2] + a[n1 - 1, n1 - 1]
                s = (a[n1 - 1, n1 - 1] - a[n1 - 2, n1 - 2]) ** 2 + 4 * a[n1 - 1, n1 - 2] * a[n1 - 2, n1 - 1]
                if s.real > 0:
                    s = sqrt(s)
                else:
                    s = sqrt(-s) * 1j
                if s.imag == 0:
                    s = s.real
                a0 = (t + s) / 2
                b0 = (t - s) / 2
                if abs(a[n1 - 1, n1 - 1] - a0) > abs(a[n1 - 1, n1 - 1] - b0):
                    shift = b0
                else:
                    shift = a0
            its += 1
            c = a[n0, n0] - shift
            s = a[n0 + 1, n0]
            v = hypot(abs(c), abs(s))
            if v == 0:
                c = 1
                s = 0
            else:
                c /= v
                s /= v
            cc = c.conjugate()
            cs = s.conjugate()
            for k in range(n0, n):
                x = a[n0, k]
                y = a[n0 + 1, k]
                a[n0, k] = cc * x + cs * y
                a[n0 + 1, k] = c * y - s * x
            for k in range(min(n1, n0 + 3)):
                x = a[k, n0]
                y = a[k, n0 + 1]
                a[k, n0] = c * x + s * y
                a[k, n0 + 1] = cc * y - cs * x
            if not isinstance(q, bool):
                for k in range(n):
                    x = q[k, n0]
                    y = q[k, n0 + 1]
                    q[k, n0] = c * x + s * y
                    q[k, n0 + 1] = cc * y - cs * x
            for j in range(n0, n1 - 2):
                c = a[j + 1, j]
                s = a[j + 2, j]
                v = hypot(abs(c), abs(s))
                if v == 0:
                    a[j + 1, j] = 0.0
                    c = 1
                    s = 0
                else:
                    a[j + 1, j] = v
                    c /= v
                    s /= v
                a[j + 2, j] = 0.0
                cc = c.conjugate()
                cs = s.conjugate()
                for k in range(j + 1, n):
                    x = a[j + 1, k]
                    y = a[j + 2, k]
                    a[j + 1, k] = cc * x + cs * y
                    a[j + 2, k] = c * y - s * x
                for k in range(0, min(n1, j + 4)):
                    x = a[k, j + 1]
                    y = a[k, j + 2]
                    a[k, j + 1] = c * x + s * y
                    a[k, j + 2] = cc * y - cs * x
                if not isinstance(q, bool):
                    for k in range(0, n):
                        x = q[k, j + 1]
                        y = q[k, j + 2]
                        q[k, j + 1] = c * x + s * y
                        q[k, j + 2] = cc * y - cs * x
    e = Matrix([[a[i, i] if j == i else 0 for j in range(n)] for i in range(n)])
    er = identity(n)
    m = 1
    for i in range(1, n):
        s = a[i, i]
        for j in range(i - 1, -1, -1):
            r = 0
            for k in range(j + 1, i + 1):
                r += a[j, k] * er[k, i]
            t = a[j, j] - s
            r = -r / t
            er[j, i] = r
            m = max(m, abs(r))
        if m != 1:
            for k in range(0, i + 1):
                er[k, i] /= m
    q = q @ er
    return e, q


def svd(matrix: Matrix) -> tuple:
    print("\nWarning: The SVD function may have calculation errors\n")
    e0, u = eigen(matrix @ matrix.t())
    u = Matrix([_[::-1] for _ in u])
    e1, v = eigen(matrix.t() @ matrix)
    sigma = Matrix([[((e0[i, i] + e1[i, i]) / 2) ** 0.5 if j == i else 0 for j in range(matrix.cols - 1, -1, -1)]
                    for i in range(matrix.rows - 1, -1, -1)])
    return u, sigma, v.t()


del Array



from .matrices import Matrix
from .types import ShapeError, Union, real


class Quaternion:
    def __init__(self, w: real, x: real, y: real, z: real):
        self.__quaternion = w, x, y, z
        self.__w, self.__x, self.__y, self.__z = self.__quaternion

    def __repr__(self):
        return "({}+{}i+{}j+{}k)".format(*self.__quaternion)

    def __add__(self, other: "Quaternion"):
        return Quaternion(*list(map(lambda i, j: i + j, self.__quaternion, other.__quaternion)))

    def __sub__(self, other: "Quaternion"):
        return Quaternion(*list(map(lambda i, j: i - j, self.__quaternion, other.__quaternion)))

    def __mul__(self, other: "int | float | Quaternion"):
        if isinstance(other, Quaternion):
            w = self.__w * other.__w - self.__x * other.__x - self.__y * other.__y - self.__z * other.__z
            x = self.__w * other.__x + self.__x * other.__w + self.__y * other.__z - self.__z * other.__y
            y = self.__w * other.__y + self.__y * other.__w + self.__z * other.__x - self.__x * other.__z
            z = self.__w * other.__z + self.__z * other.__w + self.__x * other.__y - self.__y * other.__x
            return Quaternion(w, x, y, z)
        elif isinstance(other, (int, float)):
            return Quaternion(*[other * i for i in self.__quaternion])
        else:
            raise TypeError("Operation undefined")

    def __rmul__(self, other: real):
        if isinstance(other, (int, float)):
            return self.__mul__(other)
        else:
            raise TypeError("Operation undefined")

    def __truediv__(self, other: "int | float | Quaternion"):
        if isinstance(other, Quaternion):
            divisor = sum([i ** 2 for i in other.__quaternion])
            w = self.__w / other.__w
            x = (self.__x * other.__w - self.__y * other.__z + self.__z * other.__y) / divisor
            y = (self.__y * other.__w + self.__x * other.__z - self.__z * other.__x) / divisor
            z = (self.__z * other.__w - self.__y * other.__x + self.__x * other.__y) / divisor
            return Quaternion(w, x, y, z)
        elif isinstance(other, (int, float)):
            return Quaternion(*[i / other for i in self.__quaternion])
        else:
            raise TypeError("Operation undefined")

    def __round__(self, n: int = None):
        if n is None:
            return Quaternion(*[round(i) for i in self.__quaternion])
        return Quaternion(*[round(i, n) for i in self.__quaternion])

    def __abs__(self):
        return self.norm()

    def __eq__(self, other: "Quaternion"):
        return all(list(map(lambda i, j: abs(i) == abs(j), self.__quaternion, other.__quaternion)))

    def __iadd__(self, other: "Quaternion"):
        return self + other

    def __isub__(self, other: "Quaternion"):
        return self - other

    def __imul__(self, other: "Quaternion"):
        return self * other

    def __itruediv__(self, other: "Quaternion"):
        return self / other

    def __setitem__(self, key: int, value: real):
        q = list(self.__quaternion)
        q[key] = value
        self.__quaternion = tuple(q)
        self.__w, self.__x, self.__y, self.__z = self.__quaternion

    def __pos__(self):
        return Quaternion(*self.__quaternion)

    def __neg__(self):
        return Quaternion(*[-i for i in self.__quaternion])

    def data(self):
        return list(self.__quaternion)

    def norm(self):
        return sum([i ** 2 for i in self.__quaternion]) ** 0.5

    def conjugate(self):
        return Quaternion(self.__w, -self.__x, -self.__y, -self.__z)

    def normalize(self):
        norm = self.norm()
        return Quaternion(*[i / norm for i in self.__quaternion])

    def inverse(self):
        norm = self.norm()
        return Quaternion(*[i / norm for i in self.conjugate().__quaternion])


def quat(w: real = 0, x: real = 0, y: real = 0, z: real = 0) -> Quaternion:
    return Quaternion(w, x, y, z)


class Euler:
    def __init__(self, y: real, p: real, r: real):
        self.__angles = y, p, r
        self.__y, self.__p, self.__r = self.__angles

    def __repr__(self):
        return "ypr({},{},{})".format(*self.__angles)

    def __round__(self, n: int = None):
        if n is None:
            return Quaternion(*[round(i) for i in self.__angles])
        return Quaternion(*[round(i, n) for i in self.__angles])

    def __setitem__(self, key: int, value: real):
        e = list(self.__angles)
        e[key] = value
        self.__angles = tuple(e)
        self.__y, self.__p, self.__r = self.__angles

    def data(self):
        return self.__angles


def euler(yaw: real = 0, pitch: real = 0, roll: real = 0) -> Euler:
    return Euler(yaw, pitch, roll)


def convert(data: Union[Quaternion, Matrix, Euler], to: str) -> Union[Quaternion, Matrix, Euler]:
    if to not in ["Q", "M", "E"]:
        raise ValueError("The parameter 'to' can only be 'Q', 'M', or 'E'")
    elif isinstance(data, Euler):
        if to == "Q":
            from math import cos, sin
            q = [x / 2.0 for x in data.data()]
            w = sin(q[0]) * sin(q[1]) * sin(q[2]) + cos(q[0]) * cos(q[1]) * cos(q[2])
            x = -cos(q[0]) * sin(q[1]) * sin(q[2]) + sin(q[0]) * cos(q[1]) * cos(q[2])
            y = sin(q[0]) * cos(q[1]) * sin(q[2]) + cos(q[0]) * sin(q[1]) * cos(q[2])
            z = -sin(q[0]) * sin(q[1]) * cos(q[2]) + cos(q[0]) * cos(q[1]) * sin(q[2])
            return Quaternion(w, x, y, z)
    elif isinstance(data, Matrix):
        if data.shape != [3, 3]:
            raise ShapeError("The rotation matrix must be a third-order square matrix")
        elif to == "Q":
            [[m11, m12, m13],
             [m21, m22, m23],
             [m31, m32, m33]] = data
            squares = [
                m11 + m22 + m33,
                m11 - m22 - m33,
                m22 - m11 - m33,
                m33 - m11 - m22
            ]
            max_square_sum = max(squares)
            max_index = squares.index(max_square_sum)
            sqrt_val = (max_square_sum + 1.0) ** 0.5 * 0.5
            mult = 0.25 / sqrt_val
            if max_index == 0:
                w = max_index
                x = (m23 - m32) * mult
                y = (m31 - m13) * mult
                z = (m12 - m21) * mult
            elif max_index == 1:
                x = max_index
                w = (m23 - m32) * mult
                y = (m12 + m21) * mult
                z = (m31 + m13) * mult
            elif max_index == 2:
                y = max_index
                w = (m31 - m13) * mult
                x = (m12 + m21) * mult
                z = (m23 + m32) * mult
            else:
                z = max_index
                w = (m12 - m21) * mult
                x = (m31 + m13) * mult
                y = (m23 + m32) * mult
            return Quaternion(w, x, y, z)
    elif isinstance(data, Quaternion):
        if to == "M":
            w, x, y, z = data.normalize().data()
            return Matrix([
                [1 - 2 * (y ** 2 + z ** 2), 2 * (x * y - z * w), 2 * (x * z + y * w)],
                [2 * (x * y + z * w), 1 - 2 * (x ** 2 + z ** 2), 2 * (y * z - x * w)],
                [2 * (x * z - y * w), 2 * (y * z + x * w), 1 - 2 * (x ** 2 + y ** 2)]
            ], False)
        elif to == "E":
            from math import asin, atan2
            w, x, y, z = data.normalize().data()
            t0 = 2.0 * (w * x + y * z)
            t1 = 1.0 - 2.0 * (x * x + y * y)
            yaw = atan2(t0, t1)
            t2 = 2.0 * (w * y - z * x)
            pitch = asin(t2)
            t3 = 2.0 * (w * z + x * y)
            t4 = 1.0 - 2.0 * (y * y + z * z)
            roll = atan2(t3, t4)
            return Euler(yaw, pitch, roll)
    else:
        raise TypeError("Data can only be Euler or Matrix or Quaternion")



"""
Special types collection
"""

from decimal import Decimal
from typing import *

arr = Union[list, tuple]
ite = Union[list, tuple, str]
num = Union[int, float, complex]
prec = Union[int, float, str, Decimal]
real = Union[int, float]


class ArgsError(ValueError):
    pass


class ContentError(ValueError):
    pass


class LogicError(ValueError):
    pass


class RandomError(ValueError):
    pass


class ShapeError(ValueError):
    pass


class Config:
    """
    Introduction
    ==========
    The Config class is designed to manage and configure application settings. It provides a mechanism to ensure
    consistency in configurations and prevents accidental changes by limiting attribute modification.

    Roadmap
    ==========
    The Config class is currently stable. Future updates may include new configuration options
    to support additional features.

    Usage
    ==========

    Modifying Configuration:
    ----------
    You can modify the configuration by setting attributes directly, but only with predefined boolean values.

    - config.use_latex = True  # Enable LaTeX formatting

    Reading Configuration:
    ----------
    You can read the current configuration by accessing the attributes.

    - print(config.use_unicode)  # Output: False

    Ensuring Exclusive Configuration:
    ----------
    The Config class is designed to be mutually exclusive, meaning only one configuration item can be True at a time.

    - config.use_std = True  # This will automatically set other configuration items to False

    Getting Configuration Information:
    ----------
    Use the `__repr__` method to get a string representation of the current configuration.

    - print(config)  # Output: Config(use_latex=True, use_std=False, use_unicode=False)
    """
    use_latex = False
    use_unicode = False
    use_std = True
    attributes = ("attributes",)

    def __init__(self):
        Config.attributes = tuple(attr for attr in dir(Config) if attr[0] != "_" and attr != "attributes")

    def __setattr__(self, name, value):
        if name in self.attributes:
            value = bool(value)
            if value is True:
                for attr in self.attributes:
                    if attr != name and getattr(self, attr, False) is True:
                        super().__setattr__(attr, False)
            super().__setattr__(name, value)
        else:
            raise AttributeError("Attribute '{}' is read-only and cannot be added".format(name))

    def __delattr__(self, name):
        raise AttributeError("Attribute '{}' cannot be deleted".format(name))

    def __init_subclass__(cls, **kwargs):
        raise NotImplementedError("Config cannot be subclassed")

    def __copy__(self):
        raise TypeError("Config cannot be copied")

    def __deepcopy__(self, memo):
        raise TypeError("Config cannot be copied")

    def __getattribute__(self, name):
        if name == "__class__":
            raise AttributeError("Access to '__class__' is forbidden")
        return super().__getattribute__(name)

    def __repr__(self):
        return "Config({})".format(", ".join("{}={}".format(attr, getattr(self, attr)) for attr in self.attributes))


config = Config()
del Config



from math import (atan as _atan, cos as _cos, erf as _erf, exp as _exp,
                  factorial as _factorial, gamma as _gamma, log as _log, sqrt as _sqrt)


def beta_pdf(x, a=1.0, b=1.0):
    """
    Introduction
    ==========
    Beta probability density function

    Example
    ==========
    >>> beta_pdf(0.5, 2, 2)
    1.5
    >>>
    :param x: The value at which to evaluate the PDF, must be between 0 and 1.
    :param a: The first shape parameter, must be positive.
    :param b: The second shape parameter, must be positive.
    :return: The probability density at x for the Beta distribution with parameters a and b.
    """
    if a <= 0 or b <= 0:
        raise ValueError("Shape parameters a and b must be positive")
    if x < 0 or x > 1:
        return 0
    return (x ** (a - 1) * (1 - x) ** (b - 1)) / _gamma(a) / _gamma(b) * _gamma(a + b)


def binom_pmf(k, n, p):
    """
    Introduction
    ==========
    Binomial probability mass function

    Example
    ==========
    >>> binom_pmf(2, 4, 0.5)
    0.375
    >>>
    :param k: The number of successes, must be an integer between 0 and n.
    :param n: The number of trials, must be a positive integer.
    :param p: The probability of success in a single trial, must be between 0 and 1.
    :return: The probability of k successes in n trials with probability p of success in each trial.
    """
    if not 0 <= p <= 1:
        raise ValueError("Probability p must be between 0 and 1")
    if not 0 <= k <= n:
        raise ValueError("k must be between 0 and n")
    if not isinstance(n, int) or not isinstance(k, int):
        raise ValueError("n and k must be integers")
    return _factorial(n) / (_factorial(k) * _factorial(n - k)) * (p ** k) * ((1 - p) ** (n - k))


def cauchy_pdf(x, x0, gamma):
    """
    Introduction
    ==========
    Cauchy probability density function

    Example
    =========
    >>> cauchy_pdf(0, 0, 1)
    0.3183098861837907
    >>>
    :param x: The value at which to evaluate the PDF.
    :param x0: The location parameter of the distribution.
    :param gamma: The scale parameter of the distribution, must be positive.
    :return: The probability density at x for the Cauchy distribution with location parameter x0
        and scale parameter gamma.
    """
    if gamma <= 0:
        raise ValueError("Scale parameter gamma must be positive")
    return 1 / (3.141592653589793 * gamma * (1 + ((x - x0) / gamma) ** 2))


def cauchy_cdf(x, x0, gamma):
    """
    Cauchy cumulative distribution function

    Example
    =========
    >>> cauchy_cdf(0, 0, 1)
    0.5
    >>>
    :param x: The value at which to evaluate the CDF.
    :param x0: The location parameter of the distribution.
    :param gamma: The scale parameter of the distribution, must be positive.
    :return: The cumulative probability at x for the Cauchy distribution with location parameter x0
        and scale parameter gamma.
    """
    if gamma <= 0:
        raise ValueError("Scale parameter gamma must be positive")
    return 0.5 + _atan((x - x0) / gamma) / 3.141592653589793


def chi2_pdf(x, df=1):
    """
    Introduction
    ==========
    Chi-squared probability density function

    Example
    ==========
    >>> chi2_pdf(2, 2)
    0.18393972058572117
    >>>
    :param x: The value at which to evaluate the PDF, must be non-negative.
    :param df: The degrees of freedom, must be positive.
    :return: The probability density at x for the Chi-squared distribution with df degrees of freedom.
    """
    if df <= 0:
        raise ValueError("Degrees of freedom must be positive")
    if x < 0:
        return 0
    try:
        return (x ** (df / 2 - 1) * _exp(-x / 2)) / (2 ** (df / 2) * _gamma(df / 2))
    except ZeroDivisionError:
        return 0


def chi2_cdf(x, df=1):
    """
    Introduction
    ==========
    Chi-squared cumulative distribution function

    Example
    ==========
    >>> chi2_cdf(2, 2)
    0.6321205588285578
    >>>
    :param x: The value at which to evaluate the CDF.
    :param df: The degrees of freedom, must be positive.
    :return: The probability density at x for the Chi-squared distribution with df degrees of freedom.
    """
    from math import gamma
    from .maths import lower_gamma
    return lower_gamma(df / 2, x / 2) / gamma(df / 2)


def expon_pdf(x, scale=1.0):
    """
    Introduction
    ==========
    Exponential probability density function

    Example
    ==========
    >>> expon_pdf(1, 1)
    0.36787944117144233
    >>>
    :param x: The value at which to evaluate the PDF, must be non-negative.
    :param scale: The scale parameter, must be positive.
    :return: The probability density at x for the Exponential distribution with the given scale parameter.
    """
    if scale <= 0:
        raise ValueError("Scale parameter must be positive")
    return 1 / scale * _exp(-x / scale) if x >= 0 else 0


def expon_cdf(x, scale=1.0):
    """
    Introduction
    ==========
    Exponential cumulative distribution function

    Example
    ==========
    >>> expon_cdf(1, 1)
    0.6321205588285577
    >>>
    :param x: The value at which to evaluate the CDF, must be non-negative.
    :param scale: The scale parameter, must be positive.
    :return: The cumulative probability at x for the Exponential distribution with the given scale parameter.
    """
    if scale <= 0:
        raise ValueError("Scale parameter must be positive")
    if x < 0:
        return 0
    return 1 - _exp(-x / scale)


def f_pdf(x, dfnum=1, dfden=1):
    """
    Introduction
    ==========
    F probability density function

    Example
    ==========
    >>> f_pdf(1.5, 2, 3)
    0.17677669529663687
    >>>
    :param x: The value at which to evaluate the PDF, must be non-negative.
    :param dfnum: The degrees of freedom of the numerator, must be positive.
    :param dfden: The degrees of freedom of the denominator, must be positive.
    :return: The probability density at x for the F distribution with dfnum and dfden degrees of freedom.
    """
    if dfnum <= 0 or dfden <= 0:
        raise ValueError("Degrees of freedom parameters must be positive")
    if x <= 0:
        return 0
    return (_gamma((dfnum + dfden) / 2) / (_gamma(dfnum / 2) * _gamma(dfden / 2))) * (dfnum / dfden) ** (
            dfnum / 2) * x ** (dfnum / 2 - 1) / (1 + dfnum * x / dfden) ** ((dfnum + dfden) / 2)


def gamma_pdf(x, shape=1.0, scale=1.0):
    """
    Introduction
    ==========
    Gamma probability density function

    Example
    ==========
    >>> gamma_pdf(2, 1, 1)
    0.1353352832366127
    >>>
    :param x: The value at which to evaluate the PDF, must be non-negative.
    :param shape: The shape parameter, must be positive.
    :param scale: The scale parameter, must be positive.
    :return: The probability density at x for the Gamma distribution with the given shape and scale parameters.
    """
    if shape <= 0 or scale <= 0:
        raise ValueError("Shape and scale parameters must be positive")
    if x < 0:
        return 0
    return (x ** (shape - 1) * _exp(-x / scale)) / (scale ** shape * _gamma(shape))


def geometric_pmf(k, p=0.5):
    """
    Introduction
    ==========
    Geometric probability mass function

    Example
    ==========
    >>> geometric_pmf(2, 0.5)
    0.25
    >>>
    :param k: The number of trials until the first success, must be a non-negative integer.
    :param p: The probability of success on each trial, must be between 0 and 1.
    :return: The probability of getting the first success on the k-th trial.
    """
    if not 0 <= p <= 1:
        raise ValueError("Probability p must be between 0 and 1")
    if k < 0:
        raise ValueError("k must be non-negative")
    return p * (1 - p) ** (k - 1)


def hypergeom_pmf(k, mg, n, nt):
    """
    Introduction
    ==========
    Hypergeometric probability mass function

    Example
    ==========
    >>> hypergeom_pmf(2, 5, 3, 2)
    0.3
    >>>
    :param k: The number of successes, must be a non-negative integer.
    :param mg: The total number of items in the population, must be a non-negative integer.
    :param n: The number of items in the sample, must be a non-negative integer less than or equal to mg.
    :param nt: The number of successes in the population, must be a non-negative integer less than or equal to mg.
    :return: The probability of drawing k successes in a sample of size n from a population
        with mg items and nt successes.
    """
    try:
        from math import comb
    except ImportError:
        from .maths import combination as comb
    if not (isinstance(k, int) and k >= 0) or not (isinstance(mg, int) and mg >= 0) or not (
            isinstance(n, int) and n >= 0) or not (isinstance(nt, int) and nt >= 0):
        raise ValueError("All parameters must be non-negative integers")
    if n > mg:
        raise ValueError("n must be less than or equal to mg")
    if k > n:
        raise ValueError("k must be less than or equal to n")
    if nt > mg:
        raise ValueError("nt must be less than or equal to mg")
    if k > nt:
        raise ValueError("k must be less than or equal to nt")
    return comb(nt, k) * comb(mg - nt, n - k) / comb(mg, n)


def invgauss_pdf(x, mu, lambda_, alpha):
    """
    Introduction
    ==========
    Inverse Gaussian (inverse normal) probability density function

    Example
    =========
    >>> invgauss_pdf(1, 1, 1, 1)
    0.3989422804014327
    >>>
    :param x: The value at which to evaluate the PDF. It must be positive.
    :param mu: The mean of the distribution, must be positive.
    :param lambda_: The shape parameter of the distribution, must be positive.
    :param alpha: The scale parameter of the distribution, must be non-negative.
    :return: The probability density at x for the inverse Gaussian distribution
        with mean mu, shape lambda_, and scale alpha.
    """
    if mu <= 0 or lambda_ <= 0 or alpha < 0:
        raise ValueError("mu, lambda_, and alpha parameters must be positive, and alpha must be non-negative")
    if x <= 0:
        raise ValueError("x must be positive")
    return _sqrt(lambda_ / (6.283185307179586 * x ** 3)) / alpha * _exp(-lambda_ * (x - mu) ** 2 / (
            2 * mu ** 2 * x)) * (1 + alpha * (x - mu) / mu ** 2) ** (-0.5)


def levy_pdf(x, c):
    """
    Introduction
    ==========
    Lévy probability density function

    Example
    =========
    >>> levy_pdf(1, 1)
    0.24197072451914337
    >>>
    :param x: The value at which to evaluate the PDF. It must be positive.
    :param c: The scale parameter of the distribution, must be positive.
    :return: The probability density at x for the Lévy distribution with scale parameter c.
    """
    if c <= 0:
        raise ValueError("Scale parameter c must be positive")
    if x <= 0:
        raise ValueError("x must be positive")
    return _sqrt(c / 6.283185307179586) * _exp(-c / (2 * x)) / (x ** 1.5)


def log_logistic_pdf(x, alpha, beta):
    """
    Introduction
    ==========
    Log-logistic probability density function

    Example
    =========
    >>> log_logistic_pdf(1, 1, 1)
    0.25
    >>>
    :param x: The value at which to evaluate the PDF. It must be positive.
    :param alpha: The scale parameter of the distribution, must be positive.
    :param beta: The shape parameter of the distribution, must be positive.
    :return: The probability density at x for the log-logistic distribution with scale parameter alpha
        and shape parameter beta.
    """
    if alpha <= 0 or beta <= 0:
        raise ValueError("alpha and beta parameters must be positive")
    if x <= 0:
        raise ValueError("x must be positive")
    return (beta / alpha) * (x / alpha) ** (beta - 1) * (1 + (x / alpha) ** beta) ** (-2)


def log_logistic_cdf(x, alpha, beta):
    """
    Introduction
    ==========
    Log-logistic cumulative distribution function

    Example
    =========
    >>> log_logistic_cdf(1, 1, 1)
    0.5906161091496412
    >>>
    :param x: The value at which to evaluate the CDF. It must be positive.
    :param alpha: The scale parameter of the distribution, must be positive.
    :param beta: The shape parameter of the distribution, must be positive.
    :return: The cumulative probability at x for the log-logistic distribution with scale parameter alpha
        and shape parameter beta.
    """
    if alpha <= 0 or beta <= 0:
        raise ValueError("alpha and beta parameters must be positive")
    if x <= 0:
        raise ValueError("x must be positive")
    return alpha / (alpha + beta * _log(1 + (x / beta) ** alpha))


def logistic_pdf(x, mu=0, s=1):
    """
    Introduction
    ==========
    Logistic probability density function

    Example
    =========
    >>> logistic_pdf(0, 0, 1)
    0.25
    >>>
    :param x: The value at which to evaluate the PDF.
    :param mu: The location parameter of the distribution.
    :param s: The scale parameter of the distribution, must be positive.
    :return: The probability density at x for the logistic distribution with location parameter mu
        and scale parameter s.
    """
    if s <= 0:
        raise ValueError("Scale parameter s must be positive")
    return _exp(-(x - mu) / s) / (s * (1 + _exp(-(x - mu) / s)) ** 2)


def logistic_cdf(x, mu=0, s=1):
    """
    Introduction
    ==========
    Logistic cumulative distribution function

    Example
    =========
    >>> logistic_cdf(0, 0, 1)
    0.5
    >>>
    :param x: The value at which to evaluate the CDF.
    :param mu: The location parameter of the distribution.
    :param s: The scale parameter of the distribution, must be positive.
    :return: The cumulative probability at x for the logistic distribution with location parameter mu
        and scale parameter s.
    """
    if s <= 0:
        raise ValueError("Scale parameter s must be positive")
    return 1 / (1 + _exp(-(x - mu) / s))


def lognorm_pdf(x, s=1, scale=1):
    """
    Introduction
    ==========
    Lognormal probability density function

    Example
    ==========
    >>> lognorm_pdf(2, 1, 1)
    0.1568740192789811
    >>>
    :param x: The value at which to evaluate the PDF, must be non-negative.
    :param s: The shape parameter, must be positive.
    :param scale: The scale parameter, must be positive.
    :return: The probability density at x for the Lognormal distribution with the given shape and scale parameters.
    """
    if s <= 0 or scale <= 0:
        raise ValueError("Shape and scale parameters must be positive")
    if x <= 0:
        return 0
    return 1 / (x * s * 2.5066282746310002) * _exp(-((_log(x) - _log(scale)) ** 2) / (2 * s ** 2))


def lognorm_cdf(x, mu, sigma):
    """
    Introduction
    ==========
    Lognormal cumulative distribution function

    Example
    ==========
    >>> lognorm_cdf(1, 0, 1)
    0.5
    >>>
    :param x: The value at which to evaluate the CDF, must be non-negative.
    :param mu: The mean of the logarithm of the distribution.
    :param sigma: The standard deviation of the logarithm of the distribution.
    :return: The cumulative probability at x for the Lognormal distribution with the given mu and sigma.
    """
    if x <= 0:
        return 0
    z = (_log(x) - mu) / sigma
    return normal_cdf(z)


def logser_pmf(k, p):
    """
    Introduction
    ==========
    Logarithmic (log-series) probability mass function

    Example
    ==========
    >>> logser_pmf(2, 0.5)
    0.045084220027780106
    >>>
    :param k: The number of failures before the first success, must be a non-negative integer.
    :param p: The probability of success on each trial, must be a positive number less than 1.
    :return: The probability of observing k failures before the first success in a series of Bernoulli trials
        with success probability p.
    """
    if not (isinstance(k, int) and k >= 0):
        raise ValueError("k must be a non-negative integer")
    if not (isinstance(p, (int, float)) and 0 < p < 1):
        raise ValueError("p must be a positive number less than 1")
    if k == 0:
        return p
    else:
        return -p * (1 - p) ** (k + 1) / (k * _log(1 - p))


def multinomial_pmf(k, n, p):
    """
    Introduction
    ==========
    Multinomial probability mass function

    Example
    =========
    >>> multinomial_pmf([2, 1, 0], 3, [0.5, 0.4, 0.1])
    0.30000000000000004
    >>>
    :param k: A list of integers representing the number of successes in each category.
        The length of k must match the length of p.
    :param n: The total number of trials.
    :param p: A list of probabilities for each category. The sum of p must be equal to 1.
    :return: The probability mass for the given number of successes in each category.
    """
    if len(k) != len(p):
        raise ValueError("The length of k and p must be the same")
    if sum(k) != n:
        raise ValueError("The sum of k values must be equal to n")
    p_sum = sum(p)
    p_normalized = [pi / p_sum for pi in p]
    prob_product = 1
    for ki, pi in zip(k, p_normalized):
        prob_product *= pi ** ki
    denominator = 1
    for ki in k:
        denominator *= _factorial(ki)
    return _factorial(n) / denominator * prob_product


def nbinom_pmf(k, n=1, p=0.5):
    """
    Introduction
    ==========
    Negative binomial probability mass function

    Example
    ==========
    >>> nbinom_pmf(1, 1, 0.5)
    0.25
    >>>
    :param k: The number of successes, must be a non-negative integer.
    :param n: The number of failures before the experiment is stopped, must be a positive integer.
    :param p: The probability of success on each trial, must be between 0 and 1.
    :return: The probability of getting k successes before n failures occur in a sequence of trials
        with probability p of success on each trial.
    """
    if not 0 <= p <= 1:
        raise ValueError("Probability p must be between 0 and 1")
    if not 0 <= k <= n:
        raise ValueError("k must be between 0 and n")
    if not isinstance(n, int) or not isinstance(k, int):
        raise ValueError("n and k must be integers")
    return _factorial(n + k - 1) / (_factorial(k) * _factorial(n)) * (p ** n) * ((1 - p) ** k)


def nhypergeom_pmf(k, m, n, r):
    """
    Introduction
    ==========
    Negative hypergeometric probability mass function

    Example
    ==========
    >>> nhypergeom_pmf(2, 10, 4, 2)
    0.3333333333333333
    >>>
    :param k: The number of failures before getting r successes, must be a non-negative integer.
    :param m: The total number of items in the population, must be a non-negative integer.
    :param n: The total sample size, must be a non-negative integer less than or equal to M.
    :param r: The number of successes, must be a non-negative integer less than or equal to n.
    :return: The probability of observing k failures before observing r successes in a sample of size n
        from a population with M items.
    """
    try:
        from math import comb
    except ImportError:
        from .maths import combination as comb
    if not (isinstance(k, int) and k >= 0) or not (isinstance(m, int) and m >= 0) or not (
            isinstance(n, int) and n >= 0) or not (isinstance(r, int) and r >= 0):
        raise ValueError("All parameters must be non-negative integers")
    if n > m:
        raise ValueError("n must be less than or equal to m")
    if k + r > n:
        raise ValueError("k + r must be less than or equal to n")
    if r > m:
        raise ValueError("r must be less than or equal to m")
    return comb(n - k - 1, r - 1) * comb(m - n, k) / comb(m, r)


def normal_pdf(x, mu=0, sigma=1):
    """
    Introduction
    ==========
    Normal (Gaussian) probability density function

    Example
    ==========
    >>> normal_pdf(0, 0, 1)
    0.3989422804014327
    >>>
    :param x: The value at which to evaluate the PDF.
    :param mu: The mean or location parameter of the distribution.
    :param sigma: The standard deviation or scale parameter of the distribution, must be positive.
    :return: The probability density at x for the normal distribution with mean mu and standard deviation sigma.
    """
    if sigma <= 0:
        raise ValueError("Sigma parameter must be positive")
    return (1 / (sigma * 2.5066282746310002)) * _exp(-((x - mu) ** 2) / (2 * sigma ** 2))


def normal_cdf(x, mu=0, sigma=1):
    """
    Introduction
    ==========
    Normal (Gaussian) cumulative distribution function

    Example
    ==========
    >>> normal_cdf(0, 0, 1)
    0.5
    >>>
    :param x: The value at which to evaluate the CDF.
    :param mu: The mean or location parameter of the distribution.
    :param sigma: The standard deviation or scale parameter of the distribution, must be positive.
    :return: The cumulative probability at x for the normal distribution with mean mu and standard deviation sigma.
    """
    if sigma <= 0:
        raise ValueError("Sigma parameter must be positive")
    return 0.5 * (1 + _erf((x - mu) / (sigma * _sqrt(2))))


def pareto_pdf(x, k, m):
    """
    Introduction
    ==========
    Pareto probability density function

    Example
    =========
    >>> pareto_pdf(10, 2, 5)
    0.00040000000000000013
    >>>
    :param x: The value at which to evaluate the PDF. It must be greater than or equal to the scale parameter m.
    :param k: The shape parameter of the distribution, must be positive.
    :param m: The scale parameter of the distribution, must be positive.
    :return: The probability density at x for the Pareto distribution with shape parameter k and scale parameter m.
    """
    if k <= 0 or m <= 0:
        raise ValueError("Both k and m parameters must be positive")
    if x < m:
        raise ValueError("x must be greater than or equal to the scale parameter m")
    return k / m * (1 / x) ** (k + 1)


def poisson_pmf(k, mu=1):
    """
    Introduction
    ==========
    Poisson probability mass function

    Example
    ==========
    >>> poisson_pmf(1, 1)
    0.36787944117144233
    >>>
    :param k: The number of successes (events occurred), must be non-negative integer.
    :param mu: The average rate of success (events occurring), must be non-negative real number.
    :return: The probability of k successes occurring in a given time period or space.
    """
    if k < 0:
        raise ValueError("k must be non-negative")
    return (mu ** k * _exp(-mu)) / _factorial(k)


def rayleigh_pdf(x, sigma):
    """
    Introduction
    ==========
    Rayleigh probability density function

    Example
    =========
    >>> rayleigh_pdf(1, 1)
    0.6065306597126334
    >>>
    :param x: The value at which to evaluate the PDF. It must be non-negative.
    :param sigma: The scale parameter of the distribution, must be positive.
    :return: The probability density at x for the Rayleigh distribution with scale parameter sigma.
    """
    if sigma <= 0:
        raise ValueError("Scale parameter sigma must be positive")
    if x < 0:
        raise ValueError("x must be non-negative")
    return x / (sigma ** 2) * _exp(-x ** 2 / (2 * sigma ** 2))


def t_pdf(x, df=1):
    """
    Introduction
    ==========
    Student's t probability density function

    Example
    ==========
    >>> t_pdf(1, 2)
    0.1924500897298753
    >>>
    :param x: The value at which to evaluate the PDF.
    :param df: The degrees of freedom, must be positive.
    :return: The probability density at x for the Student's t distribution with df degrees of freedom.
    """
    if df <= 0:
        raise ValueError("Degrees of freedom must be positive")
    return _gamma((df + 1) / 2) / (_sqrt(df * 3.141592653589793)
                                   * _gamma(df / 2)) * (1 + x ** 2 / df) ** (-(df + 1) / 2)


def uniform_pdf(x, loc=0, scale=1):
    """
    Introduction
    ==========
    Continuous uniform probability density function

    Example
    ==========
    >>> uniform_pdf(0.5, 0, 1)
    1.0
    >>>
    :param x: The value at which to evaluate the PDF.
    :param loc: The location parameter, the lower bound of the distribution.
    :param scale: The scale parameter, the width of the interval over which the distribution is defined,
        must be positive.
    :return: The probability density at x for the uniform distribution between loc and loc + scale.
    """
    if scale <= 0:
        raise ValueError("Scale parameter must be positive")
    return 1 / scale if loc <= x < loc + scale else 0


def uniform_cdf(x, loc=0, scale=1):
    """
    Introduction
    ==========
    Continuous uniform cumulative distribution function

    Example
    ==========
    >>> uniform_cdf(0.5, 0, 1)
    0.5
    >>>
    :param x: The value at which to evaluate the CDF.
    :param loc: The location parameter, the lower bound of the distribution.
    :param scale: The scale parameter, the width of the interval over which the distribution is defined,
        must be positive.
    :return: The cumulative probability at x for the uniform distribution between loc and loc + scale.
    """
    if scale <= 0:
        raise ValueError("Scale parameter must be positive")
    return 0 if x < loc else (x - loc) / scale if loc <= x <= loc + scale else 1


def vonmises_pdf(x, mu, kappa):
    """
    Introduction
    ==========
    Von Mises probability density function

    Example
    ==========
    >>> vonmises_pdf(0.7853981633974483, 1.5707963267948966, 2)
    0.28717685154709915
    >>>
    :param x: The value at which to evaluate the PDF, must be in the range [0, 2*pi).
    :param mu: The location parameter, representing the mean direction, must be in the range [0, 2*pi).
    :param kappa: The concentration parameter, must be positive.
    :return: The probability density at x for the von Mises distribution with the given parameters.
    """
    from .special import besseli0
    if not 0 <= mu < 6.283185307179586:
        raise ValueError("mu must be between 0 and 2 * pi")
    if not 0 <= x < 6.283185307179586:
        raise ValueError("x must be between 0 and 2 * pi")
    if kappa <= 0:
        raise ValueError("kappa must be positive")
    return _exp(kappa * _cos(x - mu)) / 6.283185307179586 / besseli0(kappa)


def weibull_max_pdf(x, c=1, scale=1, loc=0):
    """
    Introduction
    ==========
    Weibull maximum probability density function

    Example
    ==========
    >>> weibull_max_pdf(0, 1, 1, 0)
    1.0
    >>>
    :param x: The value at which to evaluate the PDF, must be less than or equal to loc.
    :param c: The shape parameter, must be positive.
    :param scale: The scale parameter, must be positive.
    :param loc: The location parameter, the upper bound of the distribution.
    :return: The probability density at x for the Weibull distribution with the given parameters.
    """
    if c <= 0 or scale <= 0:
        raise ValueError("Shape and scale parameters must be positive")
    if x > loc:
        return 0
    return (c / scale) * ((loc - x) / scale) ** (c - 1) * _exp(-((loc - x) / scale) ** c)


def weibull_min_pdf(x, c=1, scale=1, loc=0):
    """
    Introduction
    ==========
    Weibull minimum probability density function

    Example
    ==========
    >>> weibull_min_pdf(1, 1, 1, 0)
    0.36787944117144233
    >>>
    :param x: The value at which to evaluate the PDF, must be greater than or equal to loc.
    :param c: The shape parameter, must be positive.
    :param scale: The scale parameter, must be positive.
    :param loc: The location parameter, the lower bound of the distribution.
    :return: The probability density at x for the Weibull distribution with the given parameters.
    """
    if c <= 0 or scale <= 0:
        raise ValueError("Shape and scale parameters must be positive")
    if x < loc:
        return 0
    return (c / scale) * ((x - loc) / scale) ** (c - 1) * _exp(-((x - loc) / scale) ** c)


def zipf_pmf(k, s, n):
    """
    Introduction
    ==========
    Zipf probability mass function

    Example
    =========
    >>> zipf_pmf(1, 1, 10)
    0.34141715214740553
    >>>
    :param k: The rank at which to evaluate the PMF. It must be an integer between 1 and n, inclusive.
    :param s: The exponent parameter of the distribution, must be positive.
    :param n: The number of elements in the distribution, must be a positive integer.
    :return: The probability mass at rank k for the Zipf distribution with exponent parameter s
        and number of elements n.
    """
    if s <= 0:
        raise ValueError("Exponent parameter s must be positive")
    if n <= 0 or not isinstance(n, int):
        raise ValueError("Number of elements n must be a positive integer")
    if k <= 0 or k > n:
        raise ValueError("Rank k must be an integer between 1 and n, inclusive")
    return 1 / k ** s / sum([1 / (i ** s) for i in range(1, n + 1)])


if __name__ == "__main__":
    import doctest

    doctest.testmod(verbose=True)



from pypynum.arrays import *
from pypynum.chars import int2superscript, superscript2int, int2subscript, subscript2int
from pypynum.ciphers import *
from pypynum.equations import *
from pypynum.files import *
from pypynum.fft import *
from pypynum.maths import *
from pypynum.multiprec import *
from pypynum.numbers import *
from pypynum.special import *
from pypynum.types import Any, Callable, Dict, Iterator, List, Tuple


def 数组(数据: list = None, 检查: bool = True) -> Array:
    return Array(数据, 检查)


def 转换为列表(数据) -> list:
    return aslist(数据)


def 转换为数组(数据) -> Array:
    return asarray(数据)


def 全部填充(形状, 填充值, 返回类型=Array):
    return full(形状, 填充值, 返回类型)


def 类似形状填充(数组A, 填充值, 返回类型=Array):
    return full_like(数组A, 填充值, 返回类型)


def 全零(形状, 返回类型=Array):
    return full(形状, 0, 返回类型)


def 类似形状全零(数组A, 返回类型=Array):
    return full_like(数组A, 0, 返回类型)


def 全一(形状, 返回类型=Array):
    return full(形状, 1, 返回类型)


def 类似形状全一(数组A, 返回类型=Array):
    return full_like(数组A, 1, 返回类型)


def 填充序列(形状, 序列=None, 重复=True, 填充=0, 返回类型=Array):
    return fill(形状, 序列, 重复, 填充, 返回类型)


def 整数转上标(标准字符串: str) -> str:
    return int2superscript(标准字符串)


def 上标转整数(上标字符串: str) -> str:
    return superscript2int(上标字符串)


def 整数转下标(标准字符串: str) -> str:
    return int2subscript(标准字符串)


def 下标转整数(下标字符串: str) -> str:
    return subscript2int(下标字符串)


def base64密码(文本: str, 解密: bool = False) -> str:
    return base_64(文本, 解密)


def 阿特巴什密码(文本: str) -> str:
    return atbash(文本)


def ROT13密码(文本: str) -> str:
    return rot13(文本)


def 凯撒密码(文本: str, 移位: int, 解密: bool = False) -> str:
    return caesar(文本, 移位, 解密)


def 维吉尼亚密码(文本: str, 密钥: str, 解密: bool = False) -> str:
    return vigenere(文本, 密钥, 解密)


def 代替密码(文本: str, 替换映射: dict, 解密: bool = False) -> str:
    return substitution(文本, 替换映射, 解密)


def 莫尔斯密码(文本: str, 解密: bool = False) -> str:
    return morse(文本, 解密)


def 普莱费尔密码(文本: str, 密钥: str, 解密: bool = False) -> str:
    return playfair(文本, 密钥, 解密)


def 希尔256密码(文本: bytes, 密钥: list, 解密: bool = False) -> bytes:
    return hill256(文本, 密钥, 解密)


def RC4初始化密钥调度算法(密钥: bytes) -> list:
    return ksa(密钥)


def RC4伪随机生成算法(密钥序列: list):
    return prga(密钥序列)


def RC4密码(文本: bytes, 密钥: bytes) -> bytes:
    return rc4(文本, 密钥)


def 线性方程组(左边: list, 右边: list) -> list:
    return lin_eq(左边, 右边)


def 多项式方程(系数: list) -> list:
    return poly_eq(系数)


def 读取(文件: str) -> list:
    return read(文件)


def 写入(文件: str, *对象: object):
    return write(文件, *对象)


def 一维傅里叶变换(*数据) -> FT1D:
    return FT1D(*数据)


def Fraction转为Decimal(分数对象: Fraction, 有效位数: int) -> Decimal:
    return frac2dec(分数对象, 有效位数)


def 多精度自然常数(有效位数: int, 方法: str = "series") -> Decimal:
    return mp_e(有效位数, 方法)


def 多精度圆周率(有效位数: int, 方法: str = "chudnovsky") -> Decimal:
    return mp_pi(有效位数, 方法)


def 多精度黄金分割率(有效位数: int, 方法: str = "algebraic") -> Decimal:
    return mp_phi(有效位数, 方法)


def 多精度正弦(x: real, 有效位数: int) -> Decimal:
    return mp_sin(x, 有效位数)


def 多精度余弦(x: real, 有效位数: int) -> Decimal:
    return mp_cos(x, 有效位数)


def 多精度自然对数(真数: real, 有效位数: int, 使用内置方法: bool = True) -> Decimal:
    return mp_ln(真数, 有效位数, 使用内置方法)


def 多精度对数(真数: real, 底数: real, 有效位数: int, 使用内置方法: bool = True) -> Decimal:
    return mp_log(真数, 底数, 有效位数, 使用内置方法)


def 多精度反正切(x: real, 有效位数: int) -> Decimal:
    return mp_atan(x, 有效位数)


def 多精度方位角(y: real, x: real, 有效位数: int) -> Decimal:
    return mp_atan2(y, x, 有效位数)


def 多精度自然指数(指数: real, 有效位数: int, 使用内置方法: bool = True) -> Decimal:
    return mp_exp(指数, 有效位数, 使用内置方法)


def 多精度双曲正弦(x: real, 有效位数: int) -> Decimal:
    return mp_sinh(x, 有效位数)


def 多精度双曲余弦(x: real, 有效位数: int) -> Decimal:
    return mp_cosh(x, 有效位数)


def 多精度反正弦(x: real, 有效位数: int) -> Decimal:
    return mp_asin(x, 有效位数)


def 多精度反余弦(x: real, 有效位数: int) -> Decimal:
    return mp_acos(x, 有效位数)


def 多精度菲涅耳正弦积分(x: real, 有效位数: int) -> Decimal:
    return mp_fresnel_s(x, 有效位数)


def 多精度菲涅耳余弦积分(x: real, 有效位数: int) -> Decimal:
    return mp_fresnel_c(x, 有效位数)


def 多精度欧拉伽马(有效位数: int) -> Decimal:
    return mp_euler_gamma(有效位数)


def 多精度复数(实部: prec, 虚部: prec, 有效位数: int = 28) -> MPComplex:
    return MPComplex(实部, 虚部, 有效位数)


def 转为多精度复数(实部: Union[int, float, str, Decimal, complex, MPComplex], 虚部: prec,
                   有效位数: int = 28) -> MPComplex:
    return asmpc(实部, 虚部, 有效位数)


def 下伽玛(s: num, x: num) -> num:
    return lower_gamma(s, x)


def 上伽玛(s: num, x: num) -> num:
    return upper_gamma(s, x)


def 贝塞尔函数J0(x: num) -> num:
    return besselj0(x)


def 贝塞尔函数J1(x: num) -> num:
    return besselj1(x)


def 贝塞尔函数Jv(v: real, x: num) -> num:
    return besseljv(v, x)


def 贝塞尔函数I0(x: num) -> num:
    return besseli0(x)


def 贝塞尔函数I1(x: num) -> num:
    return besseli1(x)


def 贝塞尔函数Iv(v: real, x: num) -> num:
    return besseliv(v, x)


def 乘积和(*多个数组: List[Any]) -> float:
    return sumprod(*多个数组)


def x对数y乘积(x: float, y: float) -> float:
    return xlogy(x, y)


def 序列滚动(序列: Iterator[Any], 偏移: int) -> Iterator[Any]:
    return roll(序列, 偏移)


def y次方根(被开方数: num, 开方数: num) -> num:
    return root(被开方数, 开方数)


def 自然指数(指数: real) -> real:
    return exp(指数)


def 自然对数(真数: real) -> real:
    return ln(真数)


def 最大公约数(*args: int) -> int:
    return gcd(*args)


def 最小公倍数(*args: int) -> int:
    return lcm(*args)


def 正弦(x: real) -> real:
    return sin(x)


def 余弦(x: real) -> real:
    return cos(x)


def 正切(x: real) -> real:
    return tan(x)


def 余割(x: real) -> real:
    return csc(x)


def 正割(x: real) -> real:
    return sec(x)


def 余切(x: real) -> real:
    return cot(x)


def 反正弦(x: real) -> real:
    return asin(x)


def 反余弦(x: real) -> real:
    return acos(x)


def 反正切(x: real) -> real:
    return atan(x)


def 反余割(x: real) -> real:
    return acsc(x)


def 反正割(x: real) -> real:
    return asec(x)


def 反余切(x: real) -> real:
    return acot(x)


def 双曲正弦(x: real) -> real:
    return sinh(x)


def 双曲余弦(x: real) -> real:
    return cosh(x)


def 双曲正切(x: real) -> real:
    return tanh(x)


def 双曲余割(x: real) -> real:
    return csch(x)


def 双曲正割(x: real) -> real:
    return sech(x)


def 双曲余切(x: real) -> real:
    return coth(x)


def 反双曲正弦(x: real) -> real:
    return asinh(x)


def 反双曲余弦(x: real) -> real:
    return acosh(x)


def 反双曲正切(x: real) -> real:
    return atanh(x)


def 反双曲余割(x: real) -> real:
    return acsch(x)


def 反双曲正割(x: real) -> real:
    return asech(x)


def 反双曲余切(x: real) -> real:
    return acoth(x)


def 极差(数据: List[float]) -> float:
    return ptp(数据)


def 中位数(数据: List[float]) -> float:
    return median(数据)


def 频率统计(数据: List[Any]) -> Dict[Any, int]:
    return freq(数据)


def 众数(数据: List[Any]):
    return mode(数据)


def 平均数(数据: List[float]) -> float:
    return mean(数据)


def 几何平均数(数据: List[float]) -> float:
    return geom_mean(数据)


def 平方平均数(数据: List[float]) -> float:
    return square_mean(数据)


def 调和平均数(数据: List[float]) -> float:
    return harm_mean(数据)


def 原点矩(数据: List[float], 阶数: int) -> float:
    return raw_moment(数据, 阶数)


def 中心矩(数据: List[float], 阶数: int) -> float:
    return central_moment(数据, 阶数)


def 方差(数据: List[float], 自由度: int = 0) -> float:
    return var(数据, 自由度)


def 偏度(数据: List[float]) -> float:
    return skew(数据)


def 峰度(数据: List[float], 费希尔: bool = True) -> float:
    return kurt(数据, 费希尔)


def 标准差(数据: List[float], 自由度: int = 0) -> float:
    return std(数据, 自由度)


def 协方差(x: List[float], y: List[float], 自由度: int = 0) -> float:
    return cov(x, y, 自由度)


def 相关系数(x: List[float], y: List[float]) -> float:
    return corr_coeff(x, y)


def 判定系数(x: List[float], y: List[float]) -> float:
    return coeff_det(x, y)


def 分位数(数据: list, 分位值: float, 插值方法: str = "linear", 已排序: bool = False) -> float:
    return quantile(数据, 分位值, 插值方法, 已排序)


def 积累乘积(数据: List[float]) -> float:
    return product(数据)


def 连续加和(下界: int, 上界: int, 函数: Callable) -> float:
    return sigma(下界, 上界, 函数)


def 连续乘积(下界: int, 上界: int, 函数: Callable) -> float:
    return pi(下界, 上界, 函数)


def 导数(函数, 参数: float, 步长: float = 1e-6, *额外参数, **额外关键字参数) -> float:
    return deriv(函数, 参数, 步长, *额外参数, **额外关键字参数)


def 积分(函数, 积分开始: float, 积分结束: float, 积分点数: int = 1000000, *额外参数, **额外关键字参数) -> float:
    return integ(函数, 积分开始, 积分结束, 积分点数, *额外参数, **额外关键字参数)


def 贝塔函数(p: float, q: float) -> float:
    return beta(p, q)


def 伽玛函数(alpha: float) -> float:
    return gamma(alpha)


def 阶乘函数(n: int) -> int:
    return factorial(n)


def 排列数(总数: int, 选取数: int) -> int:
    return arrangement(总数, 选取数)


def 组合数(总数: int, 选取数: int) -> int:
    return combination(总数, 选取数)


def 黎曼函数(alpha: float) -> float:
    return zeta(alpha)


def 误差函数(x: real) -> real:
    return erf(x)


def S型函数(x: real) -> real:
    return sigmoid(x)


def 符号函数(x: num) -> num:
    return sign(x)


def 负一整数次幂(指数: int) -> int:
    return parity(指数)


def 累加和(序列: List[float]) -> List[float]:
    return cumsum(序列)


def 累乘积(序列: List[float]) -> List[float]:
    return cumprod(序列)


def 多次方根取整(被开方数: int, 开方数: int) -> int:
    return iroot(被开方数, 开方数)


def 欧拉函数(n: int) -> int:
    return totient(n)


def 模运算阶(a: int, n: int, b: int) -> int:
    return mod_order(a, n, b)


def 原根(a: int, 单个: bool = False) -> Union[int, List[int]]:
    return primitive_root(a, 单个)


def 归一化(数据: List[float], 目标: float = 1) -> List[float]:
    return normalize(数据, 目标)


def 加权平均(数据: List[float], 权重: List[float]) -> float:
    return average(数据, 权重)


def 扩展欧几里得算法(a: int, b: int) -> Tuple[int, int, int]:
    return exgcd(a, b)


def 中国剩余定理(n: List[int], a: List[int]) -> int:
    return crt(n, a)


def 平方根取整(被开方数: int) -> int:
    return isqrt(被开方数)


def 可能是平方数(n: int) -> bool:
    return is_possibly_square(n)


def 判断平方数(n: int) -> bool:
    return is_square(n)


def 整数转单词(整数: int) -> str:
    return int2words(整数)


def 字符串转整数(字符串: str) -> int:
    return str2int(字符串)


def 整数转罗马数(整数: int, 上划线: bool = True) -> str:
    return int2roman(整数, 上划线)


def 罗马数转整数(罗马数: str) -> int:
    return roman2int(罗马数)


def 浮点数转分数(数值: float, 是否带分数: bool = False, 误差: float = 1e-15) -> tuple:
    return float2fraction(数值, 是否带分数, 误差)


def 拆分浮点数字符串(字符串: str) -> tuple:
    return split_float(字符串)


def 解析浮点数字符串(字符串: str) -> tuple:
    return parse_float(字符串)



def read(file: str) -> list:
    """
    Read and parse files with the suffix ".pypynum"
    :param file: string
    :return:
    """
    exec("\n".join([
        "from .arrays import Array",
        "from .matrices import Matrix",
        "from .tensors import Tensor",
        "from .vectors import Vector",
        "from .geoms import Point, Line, Triangle, Quadrilateral, Polygon, Circle",
        "from .groups import Group",
        "from .quats import Quaternion, Euler",
        "from .polys import Polynomial"
    ]))
    suffix = ".pypynum"
    if not file.endswith(suffix):
        raise FileExistsError("The file extension can only be '{}'".format(suffix))
    result = []
    with open(file, "r") as r:
        data = [item.split("\\") for item in r.read().split("\n") if item]
    for item in data:
        try:
            result.append(eval("{}(*{})".format(item[0], item[1])))
        except (IndexError, NameError, SyntaxError):
            raise FileExistsError("Data '{}' does not support processing".format("\\".join(item)))
    return result


def write(file: str, *cls: object):
    """
    Save data to a file with the suffix ".pypynum"
    :param file: string
    :param cls: instance
    :return:
    """
    from .arrays import Array
    from .geoms import Point, Line, Triangle, Quadrilateral, Polygon, Circle
    from .groups import Group
    from .quats import Quaternion, Euler
    from .polys import Polynomial
    suffix = ".pypynum"
    if not file.endswith(suffix):
        raise FileExistsError("The file extension can only be '{}'".format(suffix))
    with open(file, "w") as w:
        for item in cls:
            _type = str(type(item))
            if "." not in _type:
                raise TypeError("Type '{}' does not support saving".format(type(item)))
            prefix = _type[_type.rfind(".") + 1:-2]
            if isinstance(item, Array):
                w.write("{}\\{}\n".format(prefix, str([item.data]).replace(" ", "")))
            elif isinstance(item, Point):
                w.write("{}\\{}\n".format(prefix, str([item.p]).replace(" ", "")))
            elif isinstance(item, Line):
                w.write("{}\\{}\n".format(prefix, str([item.a, item.b]).replace(" ", "")))
            elif isinstance(item, Triangle):
                w.write("{}\\{}\n".format(prefix, str([item.a, item.b, item.c]).replace(" ", "")))
            elif isinstance(item, Quadrilateral):
                w.write("{}\\{}\n".format(prefix, str([item.a, item.b, item.c, item.d]).replace(" ", "")))
            elif isinstance(item, Polygon):
                w.write("{}\\{}\n".format(prefix, str(item.points).replace(" ", "")))
            elif isinstance(item, Circle):
                w.write("{}\\{}\n".format(prefix, str([item.center, item.radius]).replace(" ", "")))
            elif isinstance(item, Group):
                w.write("{}\\{}\n".format(prefix, str([item.elements()]).replace(" ", "")))
            elif isinstance(item, (Quaternion, Euler)):
                w.write("{}\\{}\n".format(prefix, str(item.data()).replace(" ", "")))
            elif isinstance(item, Polynomial):
                w.write("{}\\{}\n".format(prefix, str([item.terms]).replace(" ", "")))
            else:
                raise TypeError("Type '{}' does not support saving".format(type(item)))



from decimal import Decimal, getcontext as __
from fractions import Fraction
from .types import Any, Union, prec

__context = __()
del __


def _setprec(sigfigs: int = 28):
    __context.prec = sigfigs


def frac2dec(frac: Fraction, sigfigs: int = 28) -> Decimal:
    __context.prec = sigfigs
    return Decimal(frac.numerator) / Decimal(frac.denominator)


def mp_e(sigfigs: int = 28, method: str = "series") -> Decimal:
    def series_e():
        nonlocal sigfigs
        sigfigs += 5
        __context.prec = sigfigs
        e = Decimal(0)
        one = Decimal(1)
        factorial = one
        n = 0
        term = one
        temp = None
        while e != temp:
            temp = e
            e += term
            n += 1
            factorial *= n
            term = one / factorial
        __context.prec -= 5
        return +e

    def spigot_e():
        a = [1] * (sigfigs + 1)
        e = ["2."]
        for _ in range(sigfigs):
            carry = 0
            for i in range(sigfigs, -1, -1):
                carry, a[i] = divmod(a[i] * 10 + carry, i + 2)
            e.append(str(carry))
        return Decimal("".join(e))

    if method == "series":
        return series_e()
    elif method == "spigot":
        return spigot_e()
    else:
        raise ValueError("Invalid method. Use 'series' or 'spigot'")


def mp_pi(sigfigs: int = 28, method: str = "chudnovsky") -> Decimal:
    def chudnovsky_pi():
        nonlocal sigfigs
        sigfigs += 5
        __context.prec = sigfigs
        c = 426880 * Decimal(10005).sqrt()
        m = 1
        _l = 13591409
        x = 1
        k = 6
        s = _l
        const = 14
        end = int(sigfigs / const) + 1
        for i in range(1, end):
            m = (k ** 3 - 16 * k) * m // (i ** 3)
            _l += 545140134
            x *= -262537412640768000
            s += Decimal(m * _l) / x
            k += 12
        pi = c / s
        __context.prec -= 5
        return +pi

    def bbp_pi():
        nonlocal sigfigs
        sigfigs += 5
        __context.prec = sigfigs
        d4 = Decimal(4)
        d2 = Decimal(2)
        d1 = Decimal(1)
        pi = Decimal(0)
        k = 0
        p16 = 1
        k8 = 0
        while True:
            pi += (d4 / (k8 + 1) - d2 / (k8 + 4) - d1 / (k8 + 5) - d1 / (k8 + 6)) / p16
            k += 1
            k8 += 8
            if k > sigfigs:
                break
            p16 <<= 4
        __context.prec -= 5
        return +pi

    if method == "chudnovsky":
        return chudnovsky_pi()
    elif method == "bbp":
        return bbp_pi()
    else:
        raise ValueError("Invalid method. Use 'chudnovsky' or 'bbp'")


def mp_phi(sigfigs: int = 28, method: str = "algebraic") -> Decimal:
    def algebraic_phi():
        __context.prec = sigfigs
        one = Decimal(1)
        five = Decimal(5)
        two = Decimal(2)
        return (one + five.sqrt()) / two

    def newton_phi():
        nonlocal sigfigs
        sigfigs += 5
        __context.prec = sigfigs
        one = Decimal(1)
        two = Decimal(2)
        x = Decimal("1.5")
        eps = one / Decimal(10 ** sigfigs)
        while True:
            x_new = x - (x ** 2 - x - one) / (two * x - one)
            if abs(x_new - x) < eps:
                break
            x = x_new
        __context.prec -= 5
        return +x

    if method == "algebraic":
        return algebraic_phi()
    elif method == "newton":
        return newton_phi()
    else:
        raise ValueError("Invalid method. Use 'algebraic' or 'newton'")


def mp_sin(x: prec, sigfigs: int = 28) -> Decimal:
    x = Decimal(str(x))
    sigfigs += 5
    __context.prec = sigfigs
    sin_x = Decimal(0)
    pi = mp_pi(sigfigs)
    x = x % (2 * pi)
    x_squared = x * x
    term = x
    n = 0
    temp = None
    while sin_x != temp:
        temp = sin_x
        sin_x += term
        n += 1
        term = -term * x_squared / ((2 * n) * (2 * n + 1))
    __context.prec -= 5
    return +sin_x


def mp_cos(x: prec, sigfigs: int = 28) -> Decimal:
    x = Decimal(str(x))
    sigfigs += 5
    __context.prec = sigfigs
    cos_x = Decimal(0)
    pi = mp_pi(sigfigs)
    x = x % (2 * pi)
    x_squared = x * x
    term = Decimal(1)
    n = 0
    temp = None
    while cos_x != temp:
        temp = cos_x
        cos_x += term
        n += 1
        term = -term * x_squared / ((2 * n - 1) * (2 * n))
    __context.prec -= 5
    return +cos_x


def mp_ln(x: prec, sigfigs: int = 28, builtin: bool = True) -> Decimal:
    if builtin:
        __context.prec = sigfigs
        return Decimal(str(x)).ln()
    x = Decimal(str(x))
    if x <= 0:
        raise ValueError("Natural logarithm is not defined for x <= 0")
    sigfigs += 5
    __context.prec = sigfigs
    sign = -1
    if x > 1:
        x = 1 / x
        sign = 1
    ln_x = Decimal(0)
    term = Decimal(1)
    dx = 1 - x
    k = 1
    temp = None
    while ln_x != temp:
        temp = ln_x
        term *= dx
        ln_x += term / k
        k += 1
    __context.prec -= 5
    return ln_x * sign


def mp_log(x: prec, base: prec, sigfigs: int = 28, builtin: bool = True) -> Decimal:
    sigfigs += 5
    if builtin:
        __context.prec = sigfigs
        x_dec = Decimal(str(x))
        base_dec = Decimal(base)
        log_x_base = x_dec.ln() / base_dec.ln()
    else:
        if x <= 0:
            raise ValueError("Logarithm is not defined for x <= 0")
        if base <= 0 or base == 1:
            raise ValueError("Logarithm base must be greater than 0 and not equal to 1")
        ln_x = mp_ln(x, sigfigs, False)
        ln_base = mp_ln(base, sigfigs, False)
        log_x_base = ln_x / ln_base
    __context.prec -= 5
    return +log_x_base


def mp_atan(x: prec, sigfigs: int = 28) -> Decimal:
    x = Decimal(str(x))
    if abs(x) != x:
        return -mp_atan(-x, sigfigs)
    sigfigs += 5
    __context.prec = sigfigs
    m = 0
    if x > 2:
        x = 1 / x
        flag = -1
    else:
        flag = 0
        if x > 0.5:
            x = x / (1 + (1 + x ** 2).sqrt())
            m = 2
            flag = 1
            if x > 0.5:
                x = x / (1 + (1 + x ** 2).sqrt())
                m = 4
    atan_x = 0
    term = power = x
    n = 1
    constant = -x * x
    temp = None
    while atan_x != temp:
        temp = atan_x
        atan_x += term
        power *= constant
        n += 2
        term = power / n
    pi = mp_pi(sigfigs)
    if flag:
        atan_x = atan_x * m if flag > 0 else pi - atan_x
    atan_x %= pi / 2
    __context.prec -= 5
    return +atan_x


def mp_atan2(y: prec, x: prec, sigfigs: int = 28) -> Decimal:
    pi = mp_pi(sigfigs + 5)
    if x == 0:
        if y == 0:
            raise ValueError("Cannot compute atan2(0, 0)")
        if y > 0:
            return pi / 2
        else:
            return -pi / 2
    result = mp_atan(y / x, sigfigs + 5)
    if x < 0:
        if y >= 0:
            result += pi
        else:
            result -= pi
    __context.prec = sigfigs
    return +result


def mp_exp(x: prec, sigfigs: int = 28, builtin: bool = True) -> Decimal:
    if builtin:
        __context.prec = sigfigs
        return Decimal(str(x)).exp()
    x = Decimal(str(x))
    sigfigs += 5
    __context.prec = sigfigs
    n = term = exp_x = Decimal(1)
    temp = None
    while exp_x != temp:
        temp = exp_x
        term *= x / n
        exp_x += term
        n += 1
    __context.prec -= 5
    return +exp_x


def mp_sinh(x: prec, sigfigs: int = 28) -> Decimal:
    x = Decimal(str(x))
    sinh_x = (Decimal(1) - mp_exp(-2 * x, sigfigs + 5)) / (2 * mp_exp(-x, sigfigs + 5))
    __context.prec = sigfigs
    return +sinh_x


def mp_cosh(x: prec, sigfigs: int = 28) -> Decimal:
    x = Decimal(str(x))
    cosh_x = (Decimal(1) + mp_exp(-2 * x, sigfigs + 5)) / (2 * mp_exp(-x, sigfigs + 5))
    __context.prec = sigfigs
    return +cosh_x


def mp_asin(x: prec, sigfigs: int = 28) -> Decimal:
    sigfigs += 5
    __context.prec = sigfigs
    x = Decimal(str(x))
    if x < -1 or x > 1:
        raise ValueError("Input must be within the domain of definition [-1, 1]")
    original_x = x
    if x > 0.5:
        x = (1 - x ** 2).sqrt()
    elif x < -0.5:
        x = -(1 - x ** 2).sqrt()
    asin_x = Decimal(0)
    x_squared = x * x
    numerator = x
    denominator = 1
    temp = None
    while asin_x != temp:
        temp = asin_x
        term = numerator / denominator
        asin_x += term
        denominator += 2
        numerator *= x_squared * (denominator - 2) / (denominator - 1)
    if original_x > 0.5:
        asin_x = mp_pi(sigfigs) / 2 - asin_x
    elif original_x < -0.5:
        asin_x = -mp_pi(sigfigs) / 2 - asin_x
    __context.prec -= 5
    return +asin_x


def mp_acos(x: prec, sigfigs: int = 28) -> Decimal:
    x = Decimal(str(x))
    if x < -1 or x > 1:
        raise ValueError("Input must be within the domain of definition [-1, 1]")
    acos_x = mp_pi(sigfigs + 5) / 2 - mp_asin(x, sigfigs + 5)
    __context.prec = sigfigs
    return +acos_x


def mp_fresnel_s(x: prec, sigfigs: int = 28) -> Decimal:
    x = Decimal(str(x))
    more = int(x * x) + 5
    sigfigs += more
    __context.prec = sigfigs
    s_x = Decimal(0)
    p = mp_pi(sigfigs)
    f = -x ** 4 * p ** 2 / 4
    num = p / 2
    coeff = 3
    fact = 1
    term = num / coeff
    n = 0
    temp = None
    while s_x != temp:
        temp = s_x
        s_x += term
        n += 1
        num *= f
        coeff += 4
        fact *= (2 * n) * (2 * n + 1)
        term = num / (coeff * fact)
    s_x *= x ** 3
    __context.prec -= more
    return +s_x


def mp_fresnel_c(x: prec, sigfigs: int = 28) -> Decimal:
    x = Decimal(str(x))
    more = int(x * x) + 5
    sigfigs += more
    __context.prec = sigfigs
    c_x = Decimal(0)
    f = -x ** 4 * mp_pi(sigfigs) ** 2 / 4
    num = Decimal(1)
    fact_2k = 1
    k = 0
    term = num / fact_2k
    n = 0
    temp = None
    while c_x != temp:
        temp = c_x
        c_x += term
        n += 1
        num *= f
        fact_2k *= (2 * n) * (2 * n - 1)
        k += 4
        term = num / (fact_2k + k * fact_2k)
    c_x *= x
    __context.prec -= more
    return +c_x


def mp_euler_gamma(sigfigs: int = 28) -> Decimal:
    sigfigs += 5
    __context.prec = sigfigs
    n = int(sigfigs * 0.6)
    i_n = Decimal(0)
    j_n = Decimal(0)
    k = 1
    n_power_k = n
    fact_k = Decimal(1)
    harmonic_sum = Decimal(0)
    temp = None
    while j_n != temp:
        temp = j_n
        term = (n_power_k / fact_k) ** 2
        i_n += term * harmonic_sum
        j_n += term
        n_power_k *= n
        fact_k *= k
        harmonic_sum += Decimal(1) / k
        k += 1
    gamma = i_n / j_n - Decimal(n).ln()
    __context.prec -= 5
    return +gamma


def mp_catalan(sigfigs: int = 28) -> Decimal:
    sigfigs += 5
    __context.prec = sigfigs
    sum1 = Decimal(0)
    sum2 = Decimal(0)
    k = 0
    neg_one = Decimal(-1)
    four = Decimal(4)
    one = Decimal(1)
    two = Decimal(2)
    three = Decimal(3)
    eight = Decimal(8)
    sixty_four = Decimal(64)
    three_fourths = three / four
    one_thirty_two = one / Decimal(32)
    temp = None
    while sum1 != temp:
        temp = sum1
        sum1 += (neg_one ** k / four ** k) * (
                two / (four * k + one) ** 2 - two / (four * k + two) ** 2 + one / (four * k + three) ** 2)
        k += 1
    k = 0
    temp = None
    while sum2 != temp:
        temp = sum2
        sum2 += (neg_one ** k / sixty_four ** k) * (
                eight / (four * k + one) ** 2 + four / (four * k + two) ** 2 + one / (four * k + three) ** 2)
        k += 1
    catalan = three_fourths * sum1 - one_thirty_two * sum2
    __context.prec -= 5
    return +catalan


class MPComplex:
    def __init__(self, real, imag, sigfigs=28):
        _setprec(sigfigs)
        self.real = +Decimal(str(real))
        self.imag = +Decimal(str(imag))
        self.sigfigs = sigfigs

    def __add__(self, other):
        other = asmpc(other, sigfigs=self.sigfigs)
        return MPComplex(self.real + other.real, self.imag + other.imag, self.sigfigs)

    def __sub__(self, other):
        other = asmpc(other, sigfigs=self.sigfigs)
        return MPComplex(self.real - other.real, self.imag - other.imag, self.sigfigs)

    def __mul__(self, other):
        other = asmpc(other, sigfigs=self.sigfigs)
        real = self.real * other.real - self.imag * other.imag
        imag = self.real * other.imag + self.imag * other.real
        return MPComplex(real, imag, self.sigfigs)

    def __truediv__(self, other):
        other = asmpc(other, sigfigs=self.sigfigs)
        if other.real == 0 and other.imag == 0:
            raise ZeroDivisionError("Cannot divide by zero")
        else:
            denominator = other.real ** 2 + other.imag ** 2
            real = (self.real * other.real + self.imag * other.imag) / denominator
            imag = (self.imag * other.real - self.real * other.imag) / denominator
            return MPComplex(real, imag, self.sigfigs)

    def __neg__(self):
        return MPComplex(-self.real, -self.imag, self.sigfigs)

    __radd__ = __add__
    __rmul__ = __mul__

    def conjugate(self):
        return MPComplex(self.real, -self.imag, self.sigfigs)

    def modulus(self, sigfigs=None):
        if sigfigs is not None:
            _setprec(sigfigs)
        mod = (self.real ** 2 + self.imag ** 2).sqrt()
        _setprec(self.sigfigs)
        return +mod

    def __abs__(self):
        return self.modulus()

    def __pow__(self, power, modulo=None):
        if modulo is not None:
            raise TypeError("Modulo operation is not supported for complex numbers")
        if isinstance(power, float):
            power = Decimal(str(power))
        elif isinstance(power, complex):
            power = MPComplex(str(power.real), str(power.imag), self.sigfigs)
        modulus = self.modulus(self.sigfigs + 5)
        angle = self.argument(self.sigfigs + 5)
        if isinstance(power, (int, Decimal)):
            new_modulus = modulus ** power
            new_angle = angle * power
        elif isinstance(power, MPComplex):
            ln_modulus = mp_ln(modulus, self.sigfigs + 5)
            exp_real = power.real
            exp_imag = power.imag
            new_modulus = (ln_modulus * exp_real - exp_imag * angle).exp()
            new_angle = exp_imag * ln_modulus + exp_real * angle
        else:
            raise ValueError("Exponent must be a real number, Decimal, or MPComplex instance")
        real = new_modulus * mp_cos(new_angle, self.sigfigs + 5)
        imag = new_modulus * mp_sin(new_angle, self.sigfigs + 5)
        return MPComplex(real, imag, self.sigfigs)

    def sqrt(self):
        modulus = self.modulus(self.sigfigs + 5)
        angle = self.argument(self.sigfigs + 5)
        sqrt_modulus = modulus.sqrt()
        half_angle = angle / 2
        real = sqrt_modulus * mp_cos(half_angle, self.sigfigs + 5)
        imag = sqrt_modulus * mp_sin(half_angle, self.sigfigs + 5)
        return MPComplex(real, imag, self.sigfigs)

    def cbrt(self):
        modulus = self.modulus(self.sigfigs + 5)
        angle = self.argument(self.sigfigs + 5)
        cbrt_modulus = modulus ** (Decimal(1) / 3)
        third_angle = angle / 3
        real = cbrt_modulus * mp_cos(third_angle, self.sigfigs + 5)
        imag = cbrt_modulus * mp_sin(third_angle, self.sigfigs + 5)
        return MPComplex(real, imag, self.sigfigs)

    def argument(self, sigfigs=None):
        if sigfigs is None:
            sigfigs = self.sigfigs
        return mp_atan2(self.imag, self.real, sigfigs)

    def ln(self):
        modulus = self.modulus(self.sigfigs + 5)
        angle = self.argument()
        return MPComplex(mp_ln(modulus, self.sigfigs), angle, self.sigfigs)

    def exp(self):
        e_real = self.real.exp()
        real_part = e_real * mp_cos(self.imag, self.sigfigs)
        imag_part = e_real * mp_sin(self.imag, self.sigfigs)
        return MPComplex(real_part, imag_part, self.sigfigs)

    def __str__(self):
        real_str = _remove_trailing_zeros(self.real)
        imag_str = _remove_trailing_zeros(self.imag)
        complex_str = ""
        if real_str != "0":
            complex_str += real_str
        if imag_str != "0":
            if (complex_str and imag_str[0] != "-") or (not complex_str and imag_str[0] == "-"):
                imag_sign = "+" if imag_str[0] != "-" else ""
            else:
                imag_sign = "-" if imag_str[0] == "-" else ""
                imag_str = imag_str.lstrip("+-")
            complex_str += imag_sign + imag_str + "i"
        if not complex_str:
            complex_str = "0"
        return complex_str

    def __repr__(self):
        return "MPComplex({!r}, {!r}, sigfigs={!r})".format(self.real, self.imag, self.sigfigs)

    def sin(self):
        real_part = mp_sin(self.real, self.sigfigs + 5) * mp_cosh(self.imag, self.sigfigs + 5)
        imag_part = mp_cos(self.real, self.sigfigs + 5) * mp_sinh(self.imag, self.sigfigs + 5)
        return MPComplex(real_part, imag_part, self.sigfigs)

    def cos(self):
        real_part = mp_cos(self.real, self.sigfigs + 5) * mp_cosh(self.imag, self.sigfigs + 5)
        imag_part = -mp_sin(self.real, self.sigfigs + 5) * mp_sinh(self.imag, self.sigfigs + 5)
        return MPComplex(real_part, imag_part, self.sigfigs)

    def tan(self):
        cos_z = self.cos()
        if cos_z.real == 0 and cos_z.imag == 0:
            raise ZeroDivisionError("Cosine of the complex number is zero, cannot compute tangent")
        return self.sin() / cos_z

    def cot(self):
        sin_z = self.sin()
        if sin_z.real == 0 and sin_z.imag == 0:
            raise ZeroDivisionError("Sine of the complex number is zero, cannot compute cotangent")
        return self.cos() / sin_z

    def sec(self):
        cos_z = self.cos()
        if cos_z.real == 0 and cos_z.imag == 0:
            raise ZeroDivisionError("Cosine of the complex number is zero, cannot compute secant")
        return MPComplex(1, 0, self.sigfigs) / cos_z

    def csc(self):
        sin_z = self.sin()
        if sin_z.real == 0 and sin_z.imag == 0:
            raise ZeroDivisionError("Sine of the complex number is zero, cannot compute cosecant")
        return MPComplex(1, 0, self.sigfigs) / sin_z


def _remove_trailing_zeros(value: Any) -> str:
    value_str = str(value).lower()
    if value_str == "0" or value_str.startswith(("0e", "-0e")):
        return "0"
    if "." not in value_str:
        return value_str
    if "e" in value_str:
        mantissa, exponent = value_str.split("e")
        mantissa = mantissa.rstrip("0").rstrip(".")
        return mantissa + "e" + exponent
    return value_str.rstrip("0").rstrip(".")


def asmpc(real: Union[int, float, str, Decimal, complex, MPComplex], imag: prec = 0,
          sigfigs: int = 28) -> MPComplex:
    if isinstance(real, (complex, MPComplex)):
        if imag != 0:
            raise ValueError("When 'real' is a complex number, 'imag' must be 0")
        return MPComplex(real.real, real.imag, sigfigs)
    elif isinstance(real, (int, float, str, Decimal)):
        if not isinstance(imag, (int, float, str, Decimal)):
            raise TypeError("When 'real' is an int, float, str, or Decimal, 'imag' must also be an int, float, str, "
                            "or Decimal")
        return MPComplex(real, imag, sigfigs)
    else:
        raise TypeError("Cannot convert to MPComplex: 'real' must be a complex number, int, float, str, or Decimal")



import math
from .types import arr, num, real

ConvError = ValueError("The absolute value of q must be less than 1 to ensure convergence.")


def besselj0(x: num) -> num:
    """
    Introduction
    ==========
    Calculate the Bessel function of the first kind of order 0, J_0(x).

    Example
    ==========
    >>> besselj0(1)
    0.7651976865579666
    >>>
    :param x: The argument of the Bessel function.
    :return: The Bessel function of the first kind of order 0 at x.
    """
    if x == 0:
        return 1.0
    result = 1.0
    term = 1.0
    k = 1
    x_squared = x * x
    while True:
        term *= -x_squared / (4 * k * k)
        temp = result
        result += term
        if result == temp:
            break
        k += 1
    return result


def besselj1(x: num) -> num:
    """
    Introduction
    ==========
    Calculate the Bessel function of the first kind of order 1, J_1(x).

    Example
    ==========
    >>> besselj1(1)
    0.44005058574493355
    >>>
    :param x: The argument of the Bessel function.
    :return: The Bessel function of the first kind of order 1 at x.
    """
    if x == 0:
        return 0.0
    result = 0.0
    term = 0.5
    k = 1
    x_squared = x * x
    while True:
        temp = result
        result += term
        if result == temp:
            break
        term *= -x_squared / (4 * k * (k + 1))
        k += 1
    return result * x


def besseljv(v: real, x: num) -> num:
    """
    Introduction
    ==========
    Calculate the Bessel function of the first kind of order v, J_v(x).

    Example
    ==========
    >>> besseljv(1, 1)
    0.44005058574493355
    >>>
    :param v: The order of the Bessel function.
    :param x: The argument of the Bessel function.
    :return: The Bessel function of the first kind of order v at x.
    """
    result = 0.0
    k = 0
    while True:
        try:
            term = (-1) ** k * (x / 2) ** (2 * k + v) / math.factorial(k) / math.gamma(k + v + 1)
            result += term
        except OverflowError:
            break
        except ValueError:
            pass
        k += 1
    return result


def besseli0(x: num) -> num:
    """
    Introduction
    ==========
    Calculate the modified Bessel function of the first kind of order 0, I_0(x).

    Example
    ==========
    >>> besseli0(1)
    1.2660658777520082
    >>>
    :param x: The argument of the modified Bessel function, must be non-negative.
    :return: The modified Bessel function of the first kind of order 0 at x.
    """
    result = 1.0
    term = 1.0
    k = 0
    while True:
        k += 1
        term *= (x / 2) ** 2 / (k * k)
        temp = result
        result += term
        if result == temp:
            break
    return result


def besseli1(x: num) -> num:
    """
    Introduction
    ==========
    Calculate the modified Bessel function of the first kind of order 1, I_1(x).

    Example
    ==========
    >>> besseli1(1)
    0.565159103992485
    >>>
    :param x: The argument of the modified Bessel function, must be non-negative.
    :return: The modified Bessel function of the first kind of order 1 at x.
    """
    result = x / 2
    term = x / 2
    k = 1
    while True:
        term *= (x / 2) ** 2 / (k * (k + 1))
        temp = result
        result += term
        if result == temp:
            break
        k += 1
    return result


def besseliv(v: real, x: num) -> num:
    """
    Introduction
    ==========
    Calculate the modified Bessel function of the first kind of order v, I_v(x).

    Example
    ==========
    >>> besseliv(1, 1)
    0.565159103992485
    >>>
    :param v: The order of the modified Bessel function.
    :param x: The argument of the modified Bessel function.
    :return: The modified Bessel function of the first kind of order v at x.
    """
    result = 0
    k = 0
    while True:
        try:
            result += (x / 2) ** (2 * k + v) / math.factorial(k) / math.gamma(k + v + 1)
        except OverflowError:
            break
        except ValueError:
            pass
        k += 1
    return result


def hyppfq(a: arr, b: arr, z: num) -> num:
    """
    Introduction
    ==========
    Calculate the generalized hypergeometric function pFq(a; b; z),
    which is a generalization of many named special functions.

    Example
    ==========
    >>> hyppfq([1, 2], [3], 0.5)
    1.545177444479562
    >>>
    :param a: List of numerator parameters of the hypergeometric function.
    :param b: List of denominator parameters of the hypergeometric function.
    :param z: The argument of the hypergeometric function.
    :return: The value of the hypergeometric function pFq(a; b; z).
    """
    p = len(a)
    q = len(b)
    result = 0
    numerator = [1] * p
    denominator = [1] * q
    z_power = 1
    factorial = 1
    n = 0
    while True:
        term = 1
        for i in range(p):
            term *= numerator[i]
        for j in range(q):
            term /= denominator[j]
        term *= z_power / factorial
        previous = result
        result += term
        if previous == result:
            break
        for i in range(p):
            numerator[i] *= (a[i] + n)
        for j in range(q):
            denominator[j] *= (b[j] + n)
        z_power *= z
        n += 1
        factorial *= n
    return result


def hyp0f1(b0: num, z: num) -> num:
    """
    Introduction
    ==========
    Calculate the hypergeometric function 0F1, which is a special case of the generalized hypergeometric function.

    Example
    ==========
    >>> hyp0f1(1, 0.5)
    1.5660829297563503
    >>>
    :param b0: The single parameter of the hypergeometric function.
    :param z: The argument of the hypergeometric function.
    :return: The value of the hypergeometric function 0F1(b0; z).
    """
    return hyppfq([], [b0], z)


def hyp1f1(a0: num, b0: num, z: num) -> num:
    """
    Introduction
    ==========
    Calculate the hypergeometric function 1F1, also known as the confluent hypergeometric function of the first kind.

    Example
    ==========
    >>> hyp1f1(1, 1, 1)
    2.7182818284590455
    >>>
    :param a0: The single numerator parameter of the hypergeometric function.
    :param b0: The single denominator parameter of the hypergeometric function.
    :param z: The argument of the hypergeometric function.
    :return: The value of the hypergeometric function 1F1(a0; b0; z).
    """
    return hyppfq([a0], [b0], z)


def hyp2f1(a0: num, a1: num, b0: num, z: num) -> num:
    """
    Introduction
    ==========
    Calculate the hypergeometric function 2F1, which is a common form of the generalized hypergeometric function.

    Example
    ==========
    >>> hyp2f1(1, 1, 1, 0.5)
    2.0
    >>>
    :param a0: The first numerator parameter of the hypergeometric function.
    :param a1: The second numerator parameter of the hypergeometric function.
    :param b0: The single denominator parameter of the hypergeometric function.
    :param z: The argument of the hypergeometric function.
    :return: The value of the hypergeometric function 2F1(a0, a1; b0; z).
    """
    return hyppfq([a0, a1], [b0], z)


def qpochhammer(a: num, q: num, n: num = None) -> num:
    """
    Introduction
    ==========
    Calculate the q-Pochhammer symbol (or q-shifted factorial) for given parameters.

    The q-Pochhammer symbol is defined as: (a; q)_n = (1 - a) * (1 - a*q) * (1 - a*q^2) * ... * (1 - a*q^(n-1))

    Example
    ==========
    >>> qpochhammer(2 + 1j, 0.5 + 0.1j, 2 + 1j)
    (-0.33353429405776575+1.8573191887407854j)
    >>>
    :param a: The base parameter of the q-Pochhammer symbol.
    :param q: The q-parameter, must have an absolute value less than 1 for convergence.
    :param n: The number of terms in the product. If None, it defaults to infinity. If n is a complex number with a
        non-zero imaginary part or a non-integer real part, the function returns the ratio of two q-Pochhammer symbols.
    :return: The value of the q-Pochhammer symbol.
    """
    if n is None:
        n = float("inf")
    elif isinstance(n, (float, complex)):
        if n.imag or not n.real.is_integer():
            return qpochhammer(a, q) / qpochhammer(a * q ** n, q)
        else:
            n = int(n.real)
    if n == float("inf") and abs(q) >= 1:
        raise ValueError("The absolute value of q must be less than 1 to ensure convergence for infinite terms.")
    result = 1
    k = 0
    while k < n:
        term = 1 - a * (q ** k)
        if term == 1:
            break
        previous = result
        result *= term
        if result == previous:
            break
        k += 1
    return result


def qfactorial(n: num, q: num) -> num:
    """
    Introduction
    ==========
    Calculate the q-factorial of a given number n using the q-Pochhammer symbol.

    The q-factorial is defined as: [n]_q! = (q; q)_n / (1 - q)^n

    Example
    ==========
    >>> qfactorial(2 + 1j, 0.5 + 0.1j)
    (1.0769270963525002+0.6870695642029011j)
    >>>
    :param n: The number for which to calculate the q-factorial.
    :param q: The q-parameter, must have an absolute value less than 1 for convergence.
    :return: The value of the q-factorial.
    """
    if abs(q) >= 1:
        raise ConvError
    qpoch_q = qpochhammer(q, q)
    qpoch_q_pow_n = qpochhammer(q ** (1 + n), q)
    result = qpoch_q / ((1 - q) ** n * qpoch_q_pow_n)
    return result


def qgamma(n: num, q: num) -> num:
    """
    Introduction
    ==========
    Calculate the q-gamma function for a given number n using the q-Pochhammer symbol.

    The q-gamma function is defined as: Γ_q(n) = (q; q)_∞ / (q^n; q)_∞ * (1 - q)^(1 - n)

    Example
    ==========
    >>> qgamma(2 + 1j, 0.5 + 0.1j)
    (0.7815732121286768+0.23898083786505578j)
    >>>
    :param n: The number for which to calculate the q-gamma function.
    :param q: The q-parameter, must have an absolute value less than 1 for convergence.
    :return: The value of the q-gamma function.
    """
    if abs(q) >= 1:
        raise ConvError
    if n == 0:
        return 1 / (1 - q)
    qpoch_q = qpochhammer(q, q)
    qpoch_q_pow_n = qpochhammer(q ** n, q)
    result = ((1 - q) ** (1 - n) * qpoch_q) / qpoch_q_pow_n
    return result


def qbeta(a: num, b: num, q: num) -> num:
    """
    Introduction
    ==========
    Calculate the q-beta function for given parameters using the q-gamma function.

    The q-beta function is defined as: Β_q(a, b) = Γ_q(a) * Γ_q(b) / Γ_q(a + b)

    Example
    ==========
    >>> qbeta(2, 3, 0.5)
    0.3047619047619047
    >>>
    :param a: The first parameter of the q-beta function.
    :param b: The second parameter of the q-beta function.
    :param q: The q-parameter, must have an absolute value less than 1 for convergence.
    :return: The value of the q-beta function.
    """
    if abs(q) >= 1:
        raise ConvError
    gamma_a_plus_b = qgamma(a + b, q)
    if gamma_a_plus_b == 0:
        return float("inf")
    gamma_a = qgamma(a, q)
    gamma_b = qgamma(b, q)
    return (gamma_a * gamma_b) / gamma_a_plus_b


def qbinomial(n: num, m: num, q: num) -> num:
    """
    Introduction
    ==========
    Calculate the q-binomial coefficient for given parameters using the specified formula.

    The q-binomial coefficient is defined as:
    (n choose m)_q = ((q^(m+1); q)_∞ * (q^(-m+n+1); q)_∞) / ((q; q)_∞ * (q^(n+1); q)_∞)

    Example
    ==========
    >>> qbinomial(3, 2, 0.5)
    1.75
    >>>
    :param n: The upper index of the q-binomial coefficient.
    :param m: The lower index of the q-binomial coefficient.
    :param q: The q-parameter, must have an absolute value less than 1 for convergence.
    :return: The value of the q-binomial coefficient.
    """
    if abs(q) >= 1:
        raise ConvError
    divisor = qpochhammer(q, q) * qpochhammer(q ** (n + 1), q)
    if not divisor:
        return float("inf")
    return qpochhammer(q ** (m + 1), q) * qpochhammer(q ** (-m + n + 1), q) / divisor


def qexp_small(z: num, q: num) -> num:
    """
    Introduction
    ==========
    Calculate the q-exponential function for small values of z.

    The q-exponential function is defined as: exp_q(z) = 1 / (z; q)_∞

    Example
    ==========
    >>> qexp_small(2 + 1j, 0.5 + 0.1j)
    (0.4218918210477187-1.4462561722357514j)
    >>>
    :param z: The argument of the q-exponential function.
    :param q: The q-parameter, must have an absolute value less than 1 for convergence.
    :return: The value of the q-exponential function for small z.
    """
    if abs(q) >= 1:
        raise ConvError
    qpoch_z = qpochhammer(z, q)
    if qpoch_z == 0:
        return float("inf")
    return 1 / qpoch_z


def qexp_large(z: num, q: num) -> num:
    """
    Introduction
    ==========
    Calculate the q-exponential function for large values of z.

    The q-exponential function is defined as: Exp_q(z) = (-z; q)_∞

    Example
    ==========
    >>> qexp_large(2 + 1j, 0.5 + 0.1j)
    (0.9080956334931816+11.389341006647559j)
    >>>
    :param z: The argument of the q-exponential function.
    :param q: The q-parameter, must have an absolute value less than 1 for convergence.
    :return: The value of the q-exponential function for large z.
    """
    if abs(q) >= 1:
        raise ConvError
    return qpochhammer(-z, q)


def qsin_small(x: num, q: num) -> num:
    """
    Introduction
    ==========
    Calculate the q-sine function for small values of x.

    The q-sine function is defined as: sin_q(x) = (exp_q(i*x) - exp_q(-i*x)) / (2*i*q)

    Example
    ==========
    >>> qsin_small(0.5, 0.5)
    (1.404936279305213+0j)
    >>>
    :param x: The argument of the q-sine function.
    :param q: The q-parameter, must have an absolute value less than 1 for convergence.
    :return: The value of the q-sine function for small x.
    """
    if abs(q) >= 1:
        raise ConvError
    return (qexp_small(1j * x, q) - qexp_small(-1j * x, q)) / (2 * 1j * q)


def qsin_large(x: num, q: num) -> num:
    """
    Introduction
    ==========
    Calculate the q-sine function for large values of x.

    The q-sine function is defined as: Sin_q(x) = (Exp_q(i*x) - Exp_q(-i*x)) / (2*i*q)

    Example
    ==========
    >>> qsin_large(0.5, 0.5)
    (1.904966692271702+0j)
    >>>
    :param x: The argument of the q-sine function.
    :param q: The q-parameter, must have an absolute value less than 1 for convergence.
    :return: The value of the q-sine function for large x.
    """
    if abs(q) >= 1:
        raise ConvError
    return (qexp_large(1j * x, q) - qexp_large(-1j * x, q)) / (2 * 1j * q)


def qsinh_small(x: num, q: num) -> num:
    """
    Introduction
    ==========
    Calculate the q-hyperbolic sine function for small values of x.

    The q-hyperbolic sine function is defined as: sinh_q(x) = (exp_q(x) - exp_q(-x)) / 2

    Example
    ==========
    >>> qsinh_small(0.5, 0.5)
    1.521662088829978
    >>>
    :param x: The argument of the q-hyperbolic sine function.
    :param q: The q-parameter, must have an absolute value less than 1 for convergence.
    :return: The value of the q-hyperbolic sine function for small x.
    """
    if abs(q) >= 1:
        raise ConvError
    return (qexp_small(x, q) - qexp_small(-x, q)) / 2


def qsinh_large(x: num, q: num) -> num:
    """
    Introduction
    ==========
    Calculate the q-hyperbolic sine function for large values of x.

    The q-hyperbolic sine function is defined as: Sinh_q(x) = (Exp_q(x) - Exp_q(-x)) / 2

    Example
    ==========
    >>> qsinh_large(0.5, 0.5)
    1.0477214669723844
    >>>
    :param x: The argument of the q-hyperbolic sine function.
    :param q: The q-parameter, must have an absolute value less than 1 for convergence.
    :return: The value of the q-hyperbolic sine function for large x.
    """
    if abs(q) >= 1:
        raise ConvError
    return (qexp_large(x, q) - qexp_large(-x, q)) / 2


def qcos_small(x: num, q: num) -> num:
    """
    Introduction
    ==========
    Calculate the q-cosine function for small values of x.

    The q-cosine function is defined as: cos_q(x) = (exp_q(i*x) + exp_q(-i*x)) / 2

    Example
    ==========
    >>> qcos_small(0.5, 0.5)
    (0.49401494605609575+0j)
    >>>
    :param x: The argument of the q-cosine function.
    :param q: The q-parameter, must have an absolute value less than 1 for convergence.
    :return: The value of the q-cosine function for small x.
    """
    if abs(q) >= 1:
        raise ConvError
    return (qexp_small(1j * x, q) + qexp_small(-1j * x, q)) / 2


def qcos_large(x: num, q: num) -> num:
    """
    Introduction
    ==========
    Calculate the q-cosine function for large values of x.

    The q-cosine function is defined as: Cos_q(x) = (Exp_q(i*x) + Exp_q(-i*x)) / 2

    Example
    ==========
    >>> qcos_large(0.5, 0.5)
    (0.6698396443906053+0j)
    >>>
    :param x: The argument of the q-cosine function.
    :param q: The q-parameter, must have an absolute value less than 1 for convergence.
    :return: The value of the q-cosine function for large x.
    """
    if abs(q) >= 1:
        raise ConvError
    return (qexp_large(1j * x, q) + qexp_large(-1j * x, q)) / 2


def qcosh_small(x: num, q: num) -> num:
    """
    Introduction
    ==========
    Calculate the q-hyperbolic cosine function for small values of x.

    The q-hyperbolic cosine function is defined as: cosh_q(x) = (exp_q(x) + exp_q(-x)) / 2

    Example
    ==========
    >>> qcosh_small(0.5, 0.5)
    1.9410845306250857
    >>>
    :param x: The argument of the q-hyperbolic cosine function.
    :param q: The q-parameter, must have an absolute value less than 1 for convergence.
    :return: The value of the q-hyperbolic cosine function for small x.
    """
    if abs(q) >= 1:
        raise ConvError
    return (qexp_small(x, q) + qexp_small(-x, q)) / 2


def qcosh_large(x: num, q: num) -> num:
    """
    Introduction
    ==========
    Calculate the q-hyperbolic cosine function for large values of x.

    The q-hyperbolic cosine function is defined as: Cosh_q(x) = (Exp_q(x) + Exp_q(-x)) / 2

    Example
    ==========
    >>> qcosh_large(0.5, 0.5)
    1.3365095620589869
    >>>
    :param x: The argument of the q-hyperbolic cosine function.
    :param q: The q-parameter, must have an absolute value less than 1 for convergence.
    :return: The value of the q-hyperbolic cosine function for large x.
    """
    if abs(q) >= 1:
        raise ConvError
    return (qexp_large(x, q) + qexp_large(-x, q)) / 2


def qpi(q: num) -> num:
    """
    Introduction
    ==========
    Calculate the q-pi function for a given q-parameter.

    The q-pi function is defined as: π_q = q ^ (1 / 4)([-1 / 2]_(q ^ 2)!) ^ 2

    Example
    ==========
    >>> qpi(0.5)
    1.6996350531822835
    >>>
    :param q: The q-parameter, must have an absolute value less than 1 for convergence.
    :return: The value of the q-pi function.
    """
    if abs(q) >= 1:
        raise ValueError("The absolute value of q must be less than 1 to ensure convergence.")
    return q ** 0.25 * qfactorial(-0.5, q * q) ** 2



class RBTNode:
    def __init__(self, data, color="red"):
        self.data = data
        self.color = color
        self.parent = None
        self.left = None
        self.right = None

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.data)


class RedBlackTree:
    def __init__(self):
        self.nil = RBTNode(None, "black")
        self.root = self.nil

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.root)

    def _fix_delete(self, x):
        while x != self.root and x.color == "black":
            if x == x.parent.left:
                s = x.parent.right
                if s.color == "red":
                    s.color = "black"
                    x.parent.color = "red"
                    self._left_rotate(x.parent)
                    s = x.parent.right
                if s.left.color == "black" and s.right.color == "black":
                    s.color = "red"
                    x = x.parent
                else:
                    if s.right.color == "black":
                        s.left.color = "black"
                        s.color = "red"
                        self._right_rotate(s)
                        s = x.parent.right
                    s.color = x.parent.color
                    x.parent.color = "black"
                    s.right.color = "black"
                    self._left_rotate(x.parent)
                    x = self.root
            else:
                s = x.parent.left
                if s.color == "red":
                    s.color = "black"
                    x.parent.color = "red"
                    self._right_rotate(x.parent)
                    s = x.parent.left
                if s.right.color == "black" and s.left.color == "black":
                    s.color = "red"
                    x = x.parent
                else:
                    if s.left.color == "black":
                        s.right.color = "black"
                        s.color = "red"
                        self._left_rotate(s)
                        s = x.parent.left
                    s.color = x.parent.color
                    x.parent.color = "black"
                    s.left.color = "black"
                    self._right_rotate(x.parent)
                    x = self.root
        x.color = "black"

    def _fix_insert(self, node):
        while node != self.root and node.parent.color == "red":
            if node.parent == node.parent.parent.left:
                uncle = node.parent.parent.right
                if uncle.color == "red":
                    node.parent.color = "black"
                    uncle.color = "black"
                    node.parent.parent.color = "red"
                    node = node.parent.parent
                else:
                    if node == node.parent.right:
                        node = node.parent
                        self._left_rotate(node)
                    node.parent.color = "black"
                    node.parent.parent.color = "red"
                    self._right_rotate(node.parent.parent)
            else:
                uncle = node.parent.parent.left
                if uncle.color == "red":
                    node.parent.color = "black"
                    uncle.color = "black"
                    node.parent.parent.color = "red"
                    node = node.parent.parent
                else:
                    if node == node.parent.left:
                        node = node.parent
                        self._right_rotate(node)
                    node.parent.color = "black"
                    node.parent.parent.color = "red"
                    self._left_rotate(node.parent.parent)
        self.root.color = "black"

    def _inorder_helper(self, node, result):
        stack = []
        current = node
        while stack or current != self.nil:
            if current != self.nil:
                stack.append(current)
                current = current.left
            else:
                current = stack.pop()
                result.append(current.data)
                current = current.right

    def _left_rotate(self, x):
        if x == self.nil:
            raise ValueError("Cannot rotate a nil node.")
        y = x.right
        x.right = y.left
        if y.left != self.nil:
            y.left.parent = x
        y.parent = x.parent
        if x.parent is None:
            self.root = y
        elif x == x.parent.left:
            x.parent.left = y
        else:
            x.parent.right = y
        y.left = x
        x.parent = y

    def _levelorder_helper(self):
        result = []
        if self.root == self.nil:
            return result
        queue = [self.root]
        while queue:
            node = queue.pop(0)
            result.append(node.data)
            if node.left != self.nil:
                queue.append(node.left)
            if node.right != self.nil:
                queue.append(node.right)
        return result

    def _postorder_helper(self, node, result):
        stack = []
        last_visited = self.nil
        current = node
        while stack or current != self.nil:
            if current != self.nil:
                stack.append(current)
                current = current.left
            else:
                peek_node = stack[-1]
                if peek_node.right != self.nil and last_visited != peek_node.right:
                    current = peek_node.right
                else:
                    result.append(peek_node.data)
                    last_visited = stack.pop()

    def _preorder_helper(self, node, result):
        stack = [node]
        while stack:
            current = stack.pop()
            if current != self.nil:
                result.append(current.data)
                stack.append(current.right)
                stack.append(current.left)

    def _print_tree_helper(self, node, indent, is_left):
        result = ""
        if node != self.nil:
            result += self._print_tree_helper(node.right, indent + ("│   " if is_left else "    "), False)
            result += indent + ("└── " if is_left else "┌── ") + node.data.__str__().__repr__()[1:-1] + (
                "(R)" if node.color == "red" else "(B)") + "\n"
            result += self._print_tree_helper(node.left, indent + ("    " if is_left else "│   "), True)
        return result

    def _range_search_helper(self, node, low, high, result):
        if node == self.nil:
            return
        if low < node.data:
            self._range_search_helper(node.left, low, high, result)
        if low <= node.data <= high:
            result.append(node.data)
        if high > node.data:
            self._range_search_helper(node.right, low, high, result)

    def _right_rotate(self, x):
        if x == self.nil:
            raise ValueError("Cannot rotate a nil node.")
        y = x.left
        x.left = y.right
        if y.right != self.nil:
            y.right.parent = x
        y.parent = x.parent
        if x.parent is None:
            self.root = y
        elif x == x.parent.right:
            x.parent.right = y
        else:
            x.parent.left = y
        y.right = x
        x.parent = y

    def _transplant(self, u, v):
        if u.parent is None:
            self.root = v
        elif u == u.parent.left:
            u.parent.left = v
        else:
            u.parent.right = v
        v.parent = u.parent

    def black_height(self, node=None):
        if node is None:
            node = self.root
        if node == self.nil:
            return 0
        left_height = self.black_height(node.left)
        right_height = self.black_height(node.right)
        if left_height != right_height:
            return -1
        return left_height + (1 if node.color == "black" else 0)

    def count_nodes(self, node=None):
        if node is None:
            node = self.root
        count = 0
        stack = [node]
        while stack:
            current = stack.pop()
            if current != self.nil:
                count += 1
                stack.append(current.left)
                stack.append(current.right)
        return count

    def delete(self, data):
        node_to_delete = self.root
        while node_to_delete != self.nil:
            if node_to_delete.data == data:
                break
            elif data < node_to_delete.data:
                node_to_delete = node_to_delete.left
            else:
                node_to_delete = node_to_delete.right
        if node_to_delete == self.nil:
            raise ValueError("Node to delete not found in the tree.")
        original_color = node_to_delete.color
        if node_to_delete.left == self.nil:
            x = node_to_delete.right
            self._transplant(node_to_delete, node_to_delete.right)
        elif node_to_delete.right == self.nil:
            x = node_to_delete.left
            self._transplant(node_to_delete, node_to_delete.left)
        else:
            successor = self.minimum()
            original_color = successor.color
            x = successor.right
            if successor.parent == node_to_delete:
                x.parent = successor
            else:
                self._transplant(successor, successor.right)
                successor.right = node_to_delete.right
                successor.right.parent = successor
            self._transplant(node_to_delete, successor)
            successor.left = node_to_delete.left
            successor.left.parent = successor
            successor.color = node_to_delete.color
        if original_color == "black":
            self._fix_delete(x)

    def get_depth(self, node):
        if node is None:
            return 0
        depth = 0
        while node != self.root and node != self.nil:
            node = node.parent
            depth += 1
        return depth

    def get_height(self, node):
        if node == self.nil:
            return 0
        max_height = 0
        stack = [(node, 1)]
        while stack:
            current, height = stack.pop()
            if current != self.nil:
                max_height = max(max_height, height)
                stack.append((current.left, height + 1))
                stack.append((current.right, height + 1))
        return max_height

    def get_max(self):
        return self.maximum().data

    def get_min(self):
        return self.minimum().data

    def insert(self, data):
        new_node = RBTNode(data)
        new_node.left = self.nil
        new_node.right = self.nil
        parent = None
        current = self.root
        while current != self.nil:
            parent = current
            if new_node.data < current.data:
                current = current.left
            else:
                current = current.right
        new_node.parent = parent
        if parent is None:
            self.root = new_node
        elif new_node.data < parent.data:
            parent.left = new_node
        else:
            parent.right = new_node
        new_node.color = "red"
        self._fix_insert(new_node)

    def is_red_black_tree(self):
        def is_red(node):
            return node.color == "red"

        def check_properties(node):
            if node == self.nil:
                return True, 1
            if is_red(node):
                if is_red(node.left) or is_red(node.right):
                    return False, 0
            left_valid, left_black_height = check_properties(node.left)
            right_valid, right_black_height = check_properties(node.right)
            return left_valid and right_valid and left_black_height == right_black_height, left_black_height + (
                1 if node.color == "black" else 0)

        return check_properties(self.root)[0]

    def lowest_common_ancestor(self, node1, node2):
        def lca_helper(root, a, b):
            if root == self.nil or root == a or root == b:
                return root
            left = lca_helper(root.left, a, b)
            right = lca_helper(root.right, a, b)
            if left is not None and right is not None:
                return root
            return left if left else right

        return lca_helper(self.root, node1, node2)

    def maximum(self):
        node = self.root
        while node.right != self.nil:
            node = node.right
        return node

    def minimum(self):
        node = self.root
        while node.left != self.nil:
            node = node.left
        return node

    def node_height(self, node):
        if node == self.nil:
            return -1
        max_height = -1
        stack = [(node, 0)]
        while stack:
            current, height = stack.pop()
            if current != self.nil:
                max_height = max(max_height, height)
                stack.append((current.left, height + 1))
                stack.append((current.right, height + 1))
        return max_height

    def node_rank(self, data):
        if self.search(data) is None:
            return -1
        rank = 0
        node = self.root
        while node != self.nil:
            if data < node.data:
                node = node.left
            elif data > node.data:
                rank += 1 + self.count_nodes(node.left)
                node = node.right
            else:
                rank += self.count_nodes(node.left)
                break
        return rank

    def path_length(self, data):
        length = 0
        node = self.root
        while node != self.nil:
            length += 1
            if data == node.data:
                return length
            elif data < node.data:
                node = node.left
            else:
                node = node.right
        return -1

    def path_to_node(self, data):
        path = []
        node = self.root
        while node != self.nil:
            path.append(node.data)
            if data == node.data:
                return path
            elif data < node.data:
                node = node.left
            else:
                node = node.right

    def predecessor(self, node):
        if node == self.nil:
            return None
        if node.left != self.nil:
            node = node.left
            while node.right != self.nil:
                node = node.right
            return node
        while node.parent != self.nil and node == node.parent.left:
            node = node.parent
        return node.parent

    def print_tree(self, to_console=True):
        tree_str = self._print_tree_helper(self.root, "", True)
        if to_console:
            print(tree_str)
        else:
            return tree_str

    def range_search(self, low, high):
        if low > high:
            raise ValueError("Low value must be less than or equal to high value.")
        result = []
        self._range_search_helper(self.root, low, high, result)
        return result

    def search(self, data):
        current = self.root
        while current != self.nil:
            if current.data == data:
                return current
            elif data < current.data:
                current = current.left
            else:
                current = current.right

    def subtree_maximum(self, node):
        while node.right != self.nil:
            node = node.right
        return node

    def subtree_minimum(self, node):
        while node.left != self.nil:
            node = node.left
        return node

    def subtree_node_count(self, node):
        if node == self.nil:
            return 0
        count = 0
        stack = [node]
        while stack:
            current = stack.pop()
            if current != self.nil:
                count += 1
                stack.append(current.left)
                stack.append(current.right)
        return count

    def successor(self, node):
        if node == self.nil:
            return None
        if node.right != self.nil:
            node = node.right
            while node.left != self.nil:
                node = node.left
            return node
        while node.parent != self.nil and node == node.parent.right:
            node = node.parent
        return node.parent

    def traverse(self, traversal_type="preorder"):
        result = []
        if traversal_type == "preorder":
            self._preorder_helper(self.root, result)
        elif traversal_type == "inorder":
            self._inorder_helper(self.root, result)
        elif traversal_type == "postorder":
            self._postorder_helper(self.root, result)
        elif traversal_type == "levelorder":
            self._levelorder_helper()
        else:
            raise ValueError("Invalid traversal type. Choose from 'preorder', 'inorder', 'postorder', 'levelorder'.")
        return result

    def tree_height(self):
        return self.node_height(self.root)

    def update(self, old_data, new_data):
        self.delete(old_data)
        self.insert(new_data)


class MTNode:
    def __init__(self, data):
        self.data = data
        self.children = []
        self.parent = None

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.data)

    def add_child(self, node):
        if node.parent is not None:
            node.parent.remove_child(node)
        node.parent = self
        self.children.append(node)

    def remove_child(self, node):
        if node in self.children:
            node.parent = None
            self.children.remove(node)


class MultiTree:
    def __init__(self, root=None):
        self.root = root if root is None or isinstance(root, MTNode) else MTNode(root)

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.root)

    def add_node(self, data, parent=None):
        new_node = data if isinstance(data, MTNode) else MTNode(data)
        if parent is None:
            if self.root is None:
                self.root = new_node
            else:
                raise ValueError("Root node already exists. Cannot add another root.")
        else:
            new_node.parent = parent
            parent.add_child(new_node)
        return new_node

    def apply_function(self, func):
        if self.root is None:
            return
        queue = [self.root]
        while queue:
            node = queue.pop(0)
            node.data = func(node.data)
            queue.extend(node.children)

    def clone_subtree(self, data):
        node = self.find_node(data)
        if node is None:
            return None
        cloned_root = MTNode(node.data)
        stack = [(node, cloned_root)]
        while stack:
            original, cloned = stack.pop()
            for child in original.children:
                cloned_child = MTNode(child.data)
                cloned_child.parent = cloned
                cloned.children.append(cloned_child)
                stack.append((child, cloned_child))
        return cloned_root

    def count_nodes(self):
        if self.root is None:
            return 0
        count = 0
        queue = [self.root]
        while queue:
            current_node = queue.pop(0)
            count += 1
            queue.extend(current_node.children)
        return count

    def find_leaves(self):
        if self.root is None:
            return []
        leaves = []
        queue = [self.root]
        while queue:
            current_node = queue.pop(0)
            if current_node.children is None:
                leaves.append(current_node.data)
            queue.extend(current_node.children)
        return leaves

    def find_node(self, data):
        def find_helper(node, target_data):
            if node.data == target_data:
                return node
            for child in node.children:
                result = find_helper(child, target_data)
                if result is not None:
                    return result

        if self.root is not None:
            return find_helper(self.root, data)

    def find_path(self, data):
        if self.root is None:
            return None
        path = []
        stack = [(self.root, path)]
        while stack:
            current_node, current_path = stack.pop()
            if current_node is None:
                continue
            new_path = current_path + [current_node.data]
            if current_node.data == data:
                return new_path
            for child in current_node.children:
                stack.append((child, new_path))

    def is_balanced(self):
        def check_height(node):
            if node is None:
                return 0
            children_heights = tuple(map(check_height, node.children))
            if children_heights is None:
                return 0
            if max(children_heights) - min(children_heights) > 1:
                return -1
            return 1 + max(children_heights)

        return check_height(self.root) != -1

    def node_depth(self, data):
        if self.root is None:
            return -1
        queue = [(self.root, 0)]
        while queue:
            current_node, current_depth = queue.pop(0)
            if current_node.data == data:
                return current_depth
            for child in current_node.children:
                queue.append((child, current_depth + 1))
        return -1

    def node_height(self, data):
        node = self.find_node(data)
        if node is None:
            return -1
        max_height = 0
        stack = [(node, 0)]
        while stack:
            current_node, current_height = stack.pop()
            if current_node.children is None:
                max_height = max(max_height, current_height)
            for child in current_node.children:
                stack.append((child, current_height + 1))
        return max_height

    def print_tree(self, to_console=True):
        def string(obj):
            return obj.__str__().__repr__()[1:-1]

        def print_tree_helper(node, indent, is_last):
            if node is None:
                return ""
            line = indent
            if is_last:
                line += "└── " + string(node.data)
                indent += "    "
            else:
                line += "├── " + string(node.data)
                indent += "│   "
            result = line + "\n"
            num_children = len(node.children)
            for i, child in enumerate(node.children):
                result += print_tree_helper(child, indent, i == num_children - 1)
            return result

        tree_str = print_tree_helper(self.root, "", True)
        if to_console:
            print(tree_str)
        else:
            return tree_str

    def prune_tree(self, predicate):
        if self.root is None:
            return
        queue = [(self.root, None)]
        while queue:
            node, parent = queue.pop(0)
            if predicate(node.data):
                if parent is not None:
                    parent.children.remove(node)
            else:
                queue.extend([(child, node) for child in node.children])

    def remove_node(self, data):
        def remove_helper(node, target_data):
            if node.data == target_data:
                if node == self.root:
                    self.root = None
                else:
                    node.parent.remove_child(node)
                return True
            for child in node.children:
                if remove_helper(child, target_data):
                    return True
            return False

        if self.root is None:
            raise ValueError("Tree is empty. Cannot remove node.")
        return remove_helper(self.root, data)

    def replace_node(self, old_data, new_data):
        node = self.find_node(old_data)
        if node is None:
            return False
        new_node = MTNode(new_data)
        new_node.children = node.children
        new_node.parent = node.parent
        if node.parent is not None:
            node.parent.children[node.parent.children.index(node)] = new_node
        for child in node.children:
            child.parent = new_node
        if node == self.root:
            self.root = new_node
        return True

    def search_nodes(self, predicate):
        if self.root is None:
            return []
        queue = [self.root]
        matching_nodes = []
        while queue:
            node = queue.pop(0)
            if predicate(node.data):
                matching_nodes.append(node)
            queue.extend(node.children)
        return matching_nodes

    def sort_tree(self, key_func=None):
        def default(x):
            return x

        if key_func is None:
            key_func = default
        elif not callable(key_func):
            raise ValueError("The 'key_func' parameter must be a callable function that takes a node's data as input "
                             "and returns a comparable value.")
        if self.root is None:
            return
        queue = [(self.root, 0)]
        while queue:
            node, level = queue.pop(0)
            node.children.sort(key=lambda child: key_func(child.data))
            queue.extend([(child, level + 1) for child in node.children])

    def to_list(self):
        def to_list_helper(node):
            node_list = [node.data]
            if node.children is not None:
                node_list.extend([to_list_helper(child) for child in node.children])
            return node_list

        if self.root is not None:
            return to_list_helper(self.root)
        return []

    def traverse(self, traversal_type="preorder"):
        def preorder(node):
            result = [node.data]
            for child in node.children:
                result.extend(preorder(child))
            return result

        def postorder(node):
            result = []
            for child in node.children:
                result.extend(postorder(child))
            result.append(node.data)
            return result

        def levelorder(node):
            result = []
            queue = [node]
            while queue:
                current_node = queue.pop(0)
                result.append(current_node.data)
                queue.extend(current_node.children)
            return result

        if self.root is None:
            raise ValueError("Tree is empty. Cannot perform traversal.")
        if traversal_type == "preorder":
            return preorder(self.root)
        elif traversal_type == "postorder":
            return postorder(self.root)
        elif traversal_type == "levelorder":
            return levelorder(self.root)
        else:
            raise ValueError("Invalid traversal type. Choose from 'preorder', 'postorder', 'levelorder'.")


class BTNode:
    def __init__(self, data):
        self.data = data
        self.left = None
        self.right = None
        self.parent = None

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.data)

    def add_left(self, node):
        if node.parent is not None:
            node.parent.remove_child(node)
        node.parent = self
        self.left = node

    def add_right(self, node):
        if node.parent is not None:
            node.parent.remove_child(node)
        node.parent = self
        self.right = node

    def remove_child(self, node):
        if self.left == node:
            self.left = None
        elif self.right == node:
            self.right = None
        if node.parent == self:
            node.parent = None


class BinaryTree:
    def __init__(self, root=None):
        self.root = root if root is None or isinstance(root, BTNode) else BTNode(root)

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.root)

    def add_left(self, parent, data):
        if parent is None:
            if self.root is None:
                self.root = BTNode(data)
                return self.root
            else:
                raise ValueError("Root node already exists. Cannot add another root.")
        elif not isinstance(parent, BTNode):
            raise ValueError("Parent must be a BTNode instance")
        new_node = BTNode(data)
        if parent.left is not None:
            parent.left.remove_child(parent.left)
        parent.left = new_node
        new_node.parent = parent
        return new_node

    def add_right(self, parent, data):
        if parent is None:
            if self.root is None:
                self.root = BTNode(data)
                return self.root
            else:
                raise ValueError("Root node already exists. Cannot add another root.")
        elif not isinstance(parent, BTNode):
            raise ValueError("Parent must be a BTNode instance")
        new_node = BTNode(data)
        if parent.right is not None:
            parent.right.remove_child(parent.right)
        parent.right = new_node
        new_node.parent = parent
        return new_node

    def apply_function(self, func):
        def apply_helper(node):
            if node is None:
                return
            node.data = func(node.data)
            apply_helper(node.left)
            apply_helper(node.right)

        apply_helper(self.root)

    def clone_subtree(self, data):
        node = self.find_node(data)
        if node is None:
            return None
        return self._clone_node(node)

    def _clone_node(self, node):
        if node is None:
            return None
        cloned_node = BTNode(node.data)
        cloned_node.left = self._clone_node(node.left)
        cloned_node.right = self._clone_node(node.right)
        if cloned_node.left is not None:
            cloned_node.left.parent = cloned_node
        if cloned_node.right is not None:
            cloned_node.right.parent = cloned_node
        return cloned_node

    def count_nodes(self):
        def count_helper(node):
            if node is None:
                return 0
            left_count = count_helper(node.left)
            right_count = count_helper(node.right)
            return 1 + left_count + right_count

        return count_helper(self.root)

    def find_leaves(self):
        def leaves_helper(node):
            if node is None:
                return []
            if node.left is None and node.right is None:
                return [node.data]
            return leaves_helper(node.left) + leaves_helper(node.right)

        return leaves_helper(self.root)

    def find_node(self, data):
        def find_helper(node):
            if node is None:
                return None
            if node.data == data:
                return node
            left_result = find_helper(node.left)
            if left_result is not None:
                return left_result
            return find_helper(node.right)

        return find_helper(self.root)

    def find_path(self, data):
        def path_helper(node, target_data, path):
            if node is None:
                return None
            new_path = path + [node.data]
            if node.data == target_data:
                return new_path
            left_result = path_helper(node.left, target_data, new_path)
            if left_result is not None:
                return left_result
            return path_helper(node.right, target_data, new_path)

        return path_helper(self.root, data, [])

    def is_balanced(self):
        def balanced_helper(node):
            if node is None:
                return 0, True
            left_height, left_balanced = balanced_helper(node.left)
            right_height, right_balanced = balanced_helper(node.right)
            return 1 + max(left_height, right_height), left_balanced and right_balanced and abs(
                left_height - right_height) <= 1

        return balanced_helper(self.root)[1]

    def node_depth(self, data):
        def depth_helper(node, target_data, depth):
            if node is None:
                return -1
            if node.data == target_data:
                return depth
            left_result = depth_helper(node.left, target_data, depth + 1)
            if left_result != -1:
                return left_result
            return depth_helper(node.right, target_data, depth + 1)

        return depth_helper(self.root, data, 0)

    def node_height(self, data):
        def height_helper(node):
            if node is None:
                return 0
            return 1 + max(height_helper(node.left), height_helper(node.right))

        result = self.find_node(data)
        if result is None:
            return -1
        return height_helper(result)

    def print_tree(self, to_console=True):
        def print_tree_helper(node, indent, is_last):
            if node is None:
                return ""
            line = indent
            if is_last:
                line += "└── " + str(node.data)
                indent += "    "
            else:
                line += "├── " + str(node.data)
                indent += "│   "
            result = line + "\n"
            if node.left is not None:
                result += print_tree_helper(node.left, indent, not node.right)
            if node.right is not None:
                result += print_tree_helper(node.right, indent, True)
            return result

        tree_str = print_tree_helper(self.root, "", True)
        if to_console:
            print(tree_str)
        else:
            return tree_str

    def prune_tree(self, predicate):
        def prune_helper(node, parent):
            if node is None:
                return
            if predicate(node.data):
                if parent is not None:
                    if parent.left == node:
                        parent.left = None
                    elif parent.right == node:
                        parent.right = None
                return
            prune_helper(node.left, node)
            prune_helper(node.right, node)

        prune_helper(self.root, None)

    def remove_node(self, data):
        def remove_helper(node, target_data, parent):
            if node is None:
                return False
            if node.data == target_data:
                if parent is not None:
                    if parent.left == node:
                        parent.left = None
                    elif parent.right == node:
                        parent.right = None
                return True
            if remove_helper(node.left, target_data, node):
                return True
            return remove_helper(node.right, target_data, node)

        if self.root is None:
            raise ValueError("Tree is empty. Cannot remove node.")
        if remove_helper(self.root, data, None):
            if self.root.left is None and self.root.right is None:
                self.root = None
            return True
        return False

    def replace_node(self, old_data, new_data):
        node = self.find_node(old_data)
        if node is None:
            return False
        new_node = BTNode(new_data)
        new_node.left = node.left
        new_node.right = node.right
        if node.left is not None:
            node.left.parent = new_node
        if node.right is not None:
            node.right.parent = new_node
        if node.parent is not None:
            if node.parent.left == node:
                node.parent.left = new_node
            elif node.parent.right == node:
                node.parent.right = new_node
        else:
            self.root = new_node
        new_node.parent = node.parent
        return True

    def search_nodes(self, predicate):
        def search_helper(node):
            if node is None:
                return []
            result = []
            if predicate(node.data):
                result.append(node)
            result.extend(search_helper(node.left))
            result.extend(search_helper(node.right))
            return result

        return search_helper(self.root)

    def sort_tree(self, key_func=None):
        def default(x):
            return x

        def sort_helper(node):
            if node is None:
                return
            sort_helper(node.left)
            sort_helper(node.right)
            if node.left and key_func(node.left.data) > key_func(node.data):
                node.left, node.data = node.data, node.left.data
            if node.right and key_func(node.right.data) < key_func(node.data):
                node.right, node.data = node.data, node.right.data

        if key_func is None:
            key_func = default
        elif not callable(key_func):
            raise ValueError("The 'key_func' parameter must be a callable function that takes a node's data as input "
                             "and returns a comparable value.")
        if self.root is None:
            return
        sort_helper(self.root)

    def to_list(self):
        def to_list_helper(node):
            if node is None:
                return []
            return [node.data, to_list_helper(node.left), to_list_helper(node.right)]

        return to_list_helper(self.root)

    def traverse(self, traversal_type="inorder"):
        def inorder(node):
            if node is None:
                return []
            return inorder(node.left) + [node.data] + inorder(node.right)

        def preorder(node):
            if node is None:
                return []
            return [node.data] + preorder(node.left) + preorder(node.right)

        def postorder(node):
            if node is None:
                return []
            return postorder(node.left) + postorder(node.right) + [node.data]

        def levelorder(node):
            if node is None:
                return []
            result = []
            queue = [node]
            while queue:
                current_node = queue.pop(0)
                result.append(current_node.data)
                if current_node.left is not None:
                    queue.append(current_node.left)
                if current_node.right is not None:
                    queue.append(current_node.right)
            return result

        if self.root is None:
            raise ValueError("Tree is empty. Cannot perform traversal.")
        if traversal_type == "inorder":
            return inorder(self.root)
        elif traversal_type == "preorder":
            return preorder(self.root)
        elif traversal_type == "postorder":
            return postorder(self.root)
        elif traversal_type == "levelorder":
            return levelorder(self.root)
        else:
            raise ValueError("Invalid traversal type. Choose from 'inorder', 'preorder', 'postorder', 'levelorder'.")
